#!/usr/bin/env python3
import os
import sys
import subprocess
import json
from distutils.dir_util import copy_tree
from shutil import copyfile, move
import yaml
import lazy_import
from sys import platform as _platform

#Pilot PLC and custom code scripts

Compiler = lazy_import.lazy_callable("pybars.Compiler")
# from pybars import Compiler

os = lazy_import.lazy_module("os")
json = lazy_import.lazy_module("json")
itertools = lazy_import.lazy_module("itertools")
string = lazy_import.lazy_module("string")
re = lazy_import.lazy_module("re")

#from collections import defaultdict
defaultdict = lazy_import.lazy_callable("collections.defaultdict")
fnmatch = lazy_import.lazy_module("fnmatch")

memregions = ['input', 'output', 'memory']
baseAddress        = 0x20000000
baseBitbandAddress = 0x22000000
alignment = 4

varsizes = {
        'BOOL': 1,
        'SINT': 1,
        'INT': 2,
        'DINT': 4,
        'LINT': 8,
        'USINT': 1,
        'UINT': 2,
        'UDINT': 4,
        'ULINT': 8,
        'BYTE': 1,
        'WORD': 2,
        'DWORD': 4,
        'LWORD': 8,
        'REAL': 4,
        'LREAL': 8,
        'TIME': 8, #??
        'DATE': 8, #??
        'DT': 8,   #??
        'TOD': 8,  #??
        'STRING': 126
        }

def iec2c_compile(iec2c_path, iecfile, temp_path):
  # resultfile=os.path.join(os.path.abspath(temp_path), 'result.txt')
  execute='{} -I {} -T {} {}'.format(os.path.join(iec2c_path, 'iec2c'), os.path.join(iec2c_path,'lib'), os.path.abspath(temp_path), iecfile)
  return os.system(execute)
  #with open(resultfile, 'r') as fin:
  #  print fin.read()

def init(config, model):
  configdef = { 'moduledefinitions': [] }
  for root, _dirnames, filenames in os.walk(os.path.dirname(os.path.realpath(__file__))):
    for filename in fnmatch.filter(filenames, 'configdefs.json'):
      with open(os.path.join(root, filename)) as obj:
        configdef['moduledefinitions'].extend( json.load(obj)['moduledefinitions'] )

  memmodules = config["modules"]
  if 'plugins' in config:
    memmodules = memmodules + config['plugins']

  #generate module config
  print('config:')
  print(config)
  modulelist = [dict(n[0], **n[1][0]) for n in map(lambda x: (x, list(filter(lambda y: x["fid"] == y["fid"],configdef["moduledefinitions"]))), memmodules) if n[1] != []]

  misalignedmodules = []
  #check if config and firmware align
  for module in model['Modules']:
    for confmod in modulelist:
      if confmod['slot'] == module['Slot'] + 1:
        if (confmod['fid'] != module['Name']):
          misalignedmodules.append(confmod['slot'])

  # generate start addresses if they are not configured
  for memregion in memregions:
    start = 0
    for module in modulelist:
      if (memregion+'address') not in module and (memregion+'size') in module and module[memregion+'size'] > 0:
        module[memregion+'address'] = start
        start = start + module[memregion+'size']

  #generate memory regions
  for memregion in memregions: map(lambda x: x.update({memregion+'bytes': range(x[memregion+'address'],x[memregion+'address']+x[memregion+'size'])}), filter(lambda x: (memregion+'address') in x and (memregion+'size') in x,modulelist))
  
  #calculate memory overlaps and sizes
  overlaps = []
  for memregion in memregions:
    for a, b in itertools.combinations(modulelist, 2): 
        if (memregion+'bytes') in a and (memregion+'bytes') in b:
            overlap = list(set(a[memregion+'bytes']).intersection(b[memregion+'bytes']))
            if overlap != []:
                overlaps.append({'elementa': a, 'elementb': b, 'memory' : memregion, 'address': overlap})

  mem = dict(map(lambda x: [x, {'region': x, 'size': 0, 'start': 0, 'absolute': 0}], memregions))
  for memregion in memregions:
    for module in modulelist:
      if (memregion+'address') in module and (memregion+'size') in module:
        size = module[memregion+'address'] + module[memregion+'size'] 
        mem[memregion]['size'] = size if size > mem[memregion]['size'] else size

  return (modulelist, overlaps, misalignedmodules, mem)

def parseplcvariables(temp_path, mem):
  global baseBitbandAddress
  varcsvfilepath=os.path.join(os.path.abspath(temp_path), 'VARIABLES.csv')
  with open(varcsvfilepath) as varcsvfile:
    content = [line.strip() for line in varcsvfile]

  mode = ''
  variables = []
  programs = []
  for line in content:
    if line.startswith('//'): 
      mode = line[2:].strip().lower()
    else:
      columns = line.split(';')
      if mode == 'variables' and len(columns) >= 5:
        varname = columns[2].split('.')
        if len(varname) > 2 and varname[0].upper() == "CONFIG":
          var = {'number': int(columns[0]), 'location': columns[1], 'resource': varname[1], 'type': columns[4]}
          if len(varname) == 3:
            var['instance'] = None
            var['name'] = varname[2]
          elif len(varname) >= 4:
            var['instance'] = varname[2]
            var['name'] = '.'.join(varname[3:])

          #add extra variable info
          var['isGlobal'] = True if var['instance'] == None else False
          var['extern'] = '{}__{}'.format(var['resource'], var['name'])
          var['isExternal'] = True if var['location'] == 'EXT' else False
          var['isString'] = True if var['type'] == 'STRING' else False
          var['plcVarName'] = '{}__{}'.format(var['resource'], var['name']) if var['isGlobal'] else '{}.{}'.format(var['instance'], var['name'])
          var['externdeclaration'] = 'extern __IEC_{}_p {};'.format(var['type'], var['extern']) if var['isGlobal'] else ''

          if var['isGlobal']:
            var['get'] = '{}.value'.format(var['plcVarName'])
          elif var['isExternal']:
            var['get'] = '(__GET_EXTERNAL_BY_REF({}))'.format(var['plcVarName'])
          else:
            var['get'] = '&(__GET_VAR({}))'.format(var['plcVarName'])

          if var['location'] == 'FB':
            if  var['isGlobal'] == True:
              programs.append(var)
          else:
            variables.append(var)

  #parse located variables
  regex = re.compile(r'__LOCATED_VAR\((?P<type>\w+),(?P<var>[^,]*),(?P<mem>\w),(?P<size>\w),(?P<byte>\d+)(,(?P<bit>\w))?\)')
  locatedvarfilepath=os.path.join(os.path.abspath(temp_path), 'LOCATED_VARIABLES.h')
  with open(locatedvarfilepath) as locatedvarfile:
    locatedvariables = [regex.match(line).groupdict() for line in locatedvarfile]

  for locvar in locatedvariables:
    locvar['externlocated'] = 'extern {0} * {1}_; extern IEC_BYTE * {2}_flags; extern {0} * {2}_changed;'.format(locvar['type'], locvar['var'].lower(), locvar['var'])
    #locvar['memoryaddress'] = 
    if locvar['mem'] == 'I':
      locvar['location'] = 'input' #from memregions var
    elif locvar['mem'] == 'Q':
      locvar['location'] = 'output' #from memregions var
    else:
      locvar['location'] = 'memory' #from memregions var

    #located memory
    memmax = int(locvar['byte']) + varsizes[locvar['type']]
    if memmax > mem[locvar['location']]['size']:
      mem[locvar['location']]['size'] = memmax

  #calc absolute memory position
  pos = 0
  for _key, value in mem.items():
    value['start'] = pos
    value['absolute'] = '0x{:X}'.format(pos + baseAddress)
    pos += value['size']

  #set variable memory positions
  for locvar in locatedvariables:
    reladdress = mem[locvar['location']]['start'] + int(locvar['byte'])
    absaddress = reladdress + baseAddress
    # print('location: {}, rel.addr. {:X}, abs.addr: {:X}'.format(locvar['var'], reladdress, absaddress))

    if locvar['bit'] != None:
      locvar['locatedaddressdeclaration'] = '{0} * {1}_ = ({0} *) 0x{2:X}; //bitband region\r#define {3} {1}_\r{0} * {3}_changed;\rIEC_BYTE * {3}_flags;\r'.format(locvar['type'], locvar['var'].lower(), baseBitbandAddress + (reladdress*32) + (int(locvar['bit'])*4), locvar['var'])
    else:
      locvar['locatedaddressdeclaration'] = '{0} * {1}_ = ({0} *) 0x{2:X};\r#define {3} {1}_\r{0} * {3}_changed;\rIEC_BYTE * {3}_flags;\r'.format(locvar['type'], locvar['var'].lower(), absaddress, locvar['var']) #TODO replace 0 with memorylocation


  locations = list(set(map(lambda x: x['location'], locatedvariables)))

  return programs, variables, locatedvariables, locations

def checkIfReplacementExists(key, module):
  if key in module:
    #TODO check if module is an array
    regex = re.compile(r"\{\{(.*?)+\}\}")
    m = regex.match(module[key])
    print(m)
    # TODO check that all replacements can be made, otherwise it cannot compile
    # that happens with wrong config file e.g.

#TODO: move to plugins
def io16_init_gen(nibble, direction, module):
  print('nibble:')
  print(nibble)
  print('direction:')
  print(direction)
  print('module:')
  print(module)

  return 'rpc_io16_{}_set_direction({}, {}); '.format(
    module['index'],
    module['codegen']['nibbles'][nibble]['selector'],
    module['codegen']['direction'][direction])

#TODO: move to plugins
def io16_read_gen(calls, module):
  if module['fid'] == 'io16' and 'config' in module and 'direction' in module['config']:
    gen = defaultdict(lambda : {})
    for k,v in module['config']['direction'].items():
      if not module['codegen']['nibbles'][k][v]['offset'] in gen[v]:
        gen[v][module['codegen']['nibbles'][k][v]['offset']] = []
      gen[v][module['codegen']['nibbles'][k][v]['offset']].append({ 'selector': module['codegen']['nibbles'][k]['selector'],  
        'data': module['codegen']['nibbles'][k][v] })

    #generate inputs
    if 'in' in gen:
      for offset in gen['in']:
        nibbleCalls = []
        for nibble in gen['in'][offset]:
          nibbleCalls.append("(rpc_io16_{}_get_byte({}) >> {})".format(
          module['index'],
          nibble['data']['register'],
          nibble['data']['shift']))
        calls['read'].append("*((unsigned char *)({} + {})) = {}; ".format(
                              module['input'+'absoluteaddress'], offset,
                              ' | '.join(nibbleCalls)))

    #generate outputs
    if 'out' in gen:
      for offset in gen['out']:
        nibbleCalls = []
        registerlist = []
        for nibble in gen['out'][offset]:
          if nibble['data']['register'] not in registerlist:
            calls['write'].append("rpc_io16_{}_set_byte({}, (*((unsigned char *)({} + {}))) >> {}); ".format(
                  module['index'],
                  nibble['data']['register'],
                  module['output'+'absoluteaddress'], offset, nibble['data']['shift']))
            registerlist.append(nibble['data']['register'])
                              
  return calls

def helper_markdown(this, options):
  md = options['fn'](this)
  # md - is a raw markdown string. You could handle it with MD-compiler
  # or just forward it to the result "as is"
  return md

def parsemodules(mem, modules):
  global baseAddress

  for module in modules:
    module['index'] = int(module['slot'])-1

  calls = {'read': [], 'write': [], 'init': [], 'include': []}
  compiler = Compiler()

  for key, value in mem.items():
    for module in modules:
        if (key+'address') in module:
          module[key+'absoluteaddress'] = '0x{:X}'.format(module[key+'address'] + value['start'] + baseAddress)

  for key, value in calls.items():
    for module in modules:
      if key in module and module[key] != '':
        checkIfReplacementExists(key, module)
        template = compiler.compile(module[key])
        calls[key].append(template(module))


  # module config specific calls
  for module in modules:
    if module['fid'] == 'io16' and 'config' in module and 'direction' in module['config']:
      directions = module['config']['direction']
      print('directions:')
      calls['init'].extend(list(map((lambda x: io16_init_gen(x, directions[x], module)), directions.keys())))
    calls = io16_read_gen(calls, module)

  return calls

def parsesubscriptions(subscriptions):
  subs = []
  for sub in subscriptions:
    subs.append(sub.replace('.', '__').upper() + '.flags = __IEC_SUBSCRIBE_FLAG; ')
  return subs

def parsetemplate(out_path, templatedata):
  #printable = set(string.printable)
  #filecontent = filter(lambda x: x in printable, templatefile.read())
  template_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'template')

  for _root, _dirs, files in os.walk(template_path):
    templfiles = map(lambda x: os.path.join(template_path, x) ,filter(lambda x: x.endswith('.templ'), files))

    for templfile in templfiles:
      with open(templfile) as f:
        compiler = Compiler()
        template = compiler.compile(f.read())

        output = template(templatedata)

        with open(os.path.join(out_path, os.path.splitext(os.path.basename(templfile))[0]), 'w') as f:
          f.write(output)          

def main(args, config):
  config = None

  #if _platform == "linux2":
  #  _platform = "linux"

  if args.iec2cdir == None:
    args.iec2cdir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'matiec', _platform)
  
  if not os.path.isdir(args.iec2cdir):
    print('Matiec IEC2C Compiler not found in {}. Maybe there is no compiler for your platform?'.format(args.iec2cdir))
    exit(1)

  #if no source path parameter is given try the environment var
  if args.source == None:
    try:
      source = os.path.join(args.workdir, 'basefw/stm') if args.workdir else './basefw/stm'
      if os.path.isdir(source):
        args.source = source
      else: 
        args.source = os.environ['PILOT_SOURCE']
    except:
      print("No parameter source directory given, no './basefw/stm' folder found and no PILOT_SOURCE environment variable defined, exiting")
      exit(1)

  stmmodel = {}
  modules = {}

  with open(os.path.join(os.path.abspath(args.source), 'stmmodel.json')) as stmmodelfile:
    stmmodel.update(json.load(stmmodelfile))

  #if no target path parameter is given try the environment var
  if args.target == None:
    args.target = os.path.join(args.workdir, "out") if args.workdir else os.path.join(os.getcwd(), "out")
    if not os.path.exists(args.target):
      os.makedirs(args.target)

  print("Writing output to " + args.target)

  if args.configfile != None:
    print("loading configfiles")
    with open(args.configfile) as configfile:
      if (args.configfile.endswith('.json')):
        config = json.load(configfile)
    modules, overlaps, misalignedmodules, mem = init(config, stmmodel)

    for overlap in overlaps:
      print("Overlapping memory regions! Module in slot {} and slot {} overlap in {} memory at byte position {}".format(overlap['elementa']['slot'], overlap['elementb']['slot'], overlap['memory'], ', '.join(str(s) for s in overlap['address'])))
      exit(2)

    if misalignedmodules != []:
      print("Modules in Firmware Source do not match modules in config file:")
      print("   Firmware   Config File")
      for module in stmmodel['Modules']:
        m = list(filter(lambda x: x['slot'] - 1 == module['Slot'] ,config['modules']))
        print("{0:1}: {1:10} {2:10}".format(module['Slot']+1, module['Name'], m[0]['fid'] if len(m) == 1 else 'None'))
      exit(2)

  #print(mem)
  #print(modules)

  copy_tree(args.source, args.target)
  copy_tree(os.path.join(os.path.abspath(os.path.dirname(os.path.realpath(__file__))), 'template', 'inc'), os.path.join(args.target, 'inc'))

  #additional model data
  if config and 'includes' in config:
    stmmodel['includes'] = config['includes']

  stfiles = []
  codefilearr = []
  if len(args.files) != 0 or 'ignore_files' in args:
    stfiles = list(filter(lambda x: x.endswith('.st'), args.files))
    codefilearr = list(filter(lambda x: not x.endswith('.st'), args.files))
  else:
    for file in os.listdir(os.path.join(args.workdir, 'src')):
      if file.endswith(".st"):
        stfiles.append(os.path.join(args.workdir, 'src', file))
      elif file.endswith(".c"):
        codefilearr.append(os.path.join(args.workdir, 'src', file))
  codefiles = ' '.join(codefilearr)
 
  plcfiles = ''
  if len(stfiles) > 0:
    plcfiles = 'config.h config.c resource1.c POUS.h ' # TODO - check if hardcoded files make sense here. can't use make plc with that. always use make default
    stfile = stfiles[0]
    if len(stfiles) > 1:
      stfile = os.path.join(args.target, '__program.st')
      with open(stfile, 'w') as outfile:
        for fname in stfiles:
          with open(fname) as infile:
            for line in infile:
              outfile.write(line)

    #iec2c compile and extract variables
    if iec2c_compile(args.iec2cdir, stfile, args.target) == 0:
      print('parsing variables')
      programs, variables, locatedvariables, locations = parseplcvariables(args.target, mem)

      if args.verbose:
        print(yaml.dump(mem, Dumper=yaml.Dumper, default_flow_style=False))

      calls = parsemodules(mem, modules)
      stmmodel['PLC'] = {'programs': programs, 'variables': variables, 'locatedvariables': locatedvariables, 'locations': locations, 'mem': mem, 'calls': calls}
      stmmodel['PLC']['modules'] = modules
      if 'subscriptions' in config:
        stmmodel['PLC']['subscriptions'] = parsesubscriptions(config['subscriptions'])
      parsetemplate(args.target, stmmodel)

      with open(os.path.join(args.workdir, "out", 'stmmodel.json'), 'w') as stmmodelfile:
        json.dump(stmmodel, stmmodelfile)

    else:
      print("Error executing IEC2C Compiler")
      exit(1)

  # codefiles
  for codefile in codefilearr:
    copyfile(codefile, os.path.join(args.target, os.path.basename(os.path.abspath(codefile))))

  # extra compilation files
  newMakefile = os.path.join(args.target, 'newMakefile')
  makefile = os.path.join(args.target, 'Makefile')
  with open(newMakefile, 'w') as new:
    new.write('EXCLUDE_SRCS=POUS.c\n')
    with open(makefile) as old:
      new.write(old.read())

  move(newMakefile, makefile)

  #run make - only works with installed arm-none-eabi-gcc
  #TODO - enable using docker container for the compiler
  subprocess.call(['make', '-C', args.target])

  return 0


#!/usr/bin/env python3
import os
import sys
import subprocess
import json
from distutils.dir_util import copy_tree
from shutil import copyfile, move
import yaml
import lazy_import
from sys import platform as _platform

#Pilot PLC and custom code scripts

Compiler = lazy_import.lazy_callable("pybars.Compiler")
# from pybars import Compiler

os = lazy_import.lazy_module("os")
json = lazy_import.lazy_module("json")
itertools = lazy_import.lazy_module("itertools")
string = lazy_import.lazy_module("string")
re = lazy_import.lazy_module("re")

#from collections import defaultdict
defaultdict = lazy_import.lazy_callable("collections.defaultdict")
fnmatch = lazy_import.lazy_module("fnmatch")


def init(config, model):
  pass

def program(config):
  pass

def templateparser(args, dir, model, compiler, helpers):
  for subdir, dirs, files in os.walk(dir):
    outdir = os.path.join(args.workdir, os.path.relpath(subdir, dir))
    for file in files:
      infile = os.path.join(subdir, file)
      with open(infile) as f:
        if infile.endswith('.templ'):
          template = compiler.compile(f.read())
          output = template(model, helpers)
        else:
          output = f.read()
        with open(os.path.join(outdir, os.path.splitext(os.path.basename(infile))[0]), 'w+') as f:
          f.write(output)  

def parsetemplate(out_path, templatedata, dir):
  #printable = set(string.printable)
  #filecontent = filter(lambda x: x in printable, templatefile.read())
  template_path = os.path.join(dir, 'template')
  compiler = Compiler()

  for subdir, _dirs, files in os.walk(template_path):
    outdir = os.path.join(out_path, os.path.relpath(subdir, template_path))

    for file in files:
      infile = os.path.join(subdir, file)
      with open(infile) as f:
        if infile.endswith('.templ'):
          template = compiler.compile(f.read())
          output = template(templatedata)
        #else:
        #  output = f.read()
          with open(os.path.join(outdir, os.path.splitext(os.path.basename(infile))[0]), 'w+') as f:
            f.write(output)  

def main(args, config, model, projectdir, compilerdir):
  
  parsetemplate(projectdir, model, compilerdir)
  #run make - only works with installed arm-none-eabi-gcc
  #TODO - enable using docker container for the compiler
  return subprocess.call(['make', '-C', args.workdir, '-s'])



#!/usr/bin/env python3
import sys
import argparse
import pilotsetup
import build

argparser = argparse.ArgumentParser(description='Pilot Command-Line Interface')
subparsers = argparser.add_subparsers(dest='subparser_name')

parser_a = subparsers.add_parser('setup', help="Configure Pilot Firmware")
pilotsetup.arguments(parser_a)

parser_b = subparsers.add_parser('build', help="Compile additional software into firmware")
build.arguments(parser_b)

args = argparser.parse_args()
if ('subparser_name' in args):
  if (args.subparser_name == 'setup'):
    sys.exit(pilotsetup.main(args))
  elif (args.subparser_name == 'build'):
    sys.exit(build.main(args))



import os
import distutils
import tarfile
import subprocess
import json
import yaml
from .sbc import Sbc
from .pilotdriver import PilotDriver
from .pilotserver import PilotServer
from . import helper
import importlib.util
from colorama import Fore

def download_base_firmware(args):
    modules = {}
    eeproms = {}
    detect_modules = True
    for mod in range(1, PilotDriver.MODULE_COUNT+1):
        modarg = 'm{}'.format(mod)
        if modarg in args and getattr(args, modarg) is not None:
            detect_modules = False
            eeproms[mod] = {'uid': '', 'hid': '', 'fid': getattr(args,modarg)}

    if detect_modules:
        if not args.node:
            print('We need a nodename or IP and username/password (ssh) of the Node you want to configure.')
            args.node = input('node/IP of Node to get Firmware Configuration from: ')

        with Sbc(args) as sbc:
            pilotserver = PilotServer(sbc)
            pilotdriver = PilotDriver(pilotserver, sbc)

            if args.server != None:
                pilotserver.pilot_server = args.server

            modules, success = pilotdriver.load_pilot_defs()
    else:
        pilotserver = PilotServer(None)
        pilotdriver = PilotDriver(pilotserver, None)
        modules, success = pilotdriver.getmodules(eeproms)

    if modules != None:
        for module in modules:
            print('Module {}: {}{}'.format(
                module['module'], Fore.GREEN, module['currentfid_nicename']))

    version, success = pilotdriver.get_firmware_source(os.path.join(args.workdir, 'basefw') if args.workdir else './basefw', args.fwversion)
    if not success:
        print(Fore.RED + 'Could not download firmware source!')
        exit(1)

    return version, modules

def init(args, pilotconfig_version):
    use_compiler = None
    print('This will create a new Pilot firmware project in the current folder')

    targetpath = os.path.join(args.workdir) if args.workdir else './'

    # check if the target path is empty
    if not os.path.exists(targetpath):
        try:
            os.makedirs(targetpath)
        except:
            print("Cannot create project directory. Please check permissions.")
            exit(1)
    elif len(os.listdir(targetpath)) != 0:
        print('The current folder {} is not empty.'.format(targetpath))
        print('Aborting!')
        exit(1)

    # copy default project files if pulled from github
    if not args.local:
        res = subprocess.call("git clone https://github.com/pilotnexus/pilot_plc_template.git {}".format(targetpath), shell = True)
        if res != 0:
            exit(res)

    compilers, _ = helper.get_compilers()

    if args.compiler:
        if next(x for x in compilers if x['name'] == args.compiler) == None:
            print('Could not find compiler {}. Use --show-compilers to get available compilers.'.format(args.compiler))
            return 1

    if not args.compiler:
        compiler_index = -1
        if (len(compilers) == 0):
            print('No compilers found, exiting')
            return 1
        elif (len(compilers) == 1): #only one compiler available
            compiler_index = 0

        while (compiler_index < 0 or compiler_index > len(compilers)):
            print('Please specify the compiler toolchain to use:')
            for index, compiler in enumerate(compilers):
                print('[{}] {}: {}'.format(index+1, compiler['name'], compiler['description']))
            try:
                compiler_index = int(input('[1-{}]: '.format(len(compilers)))) - 1
            except KeyboardInterrupt:
                exit(1)
            except: pass
        args.compiler = compilers[compiler_index]['name']

    try:
        firmware_version, modules = download_base_firmware(args)
        # create credentials.json
        #cred = {}
        #try:
        #  nodeconf = yaml.load(sbc.getFileContent('/etc/pilot/pilotnode.yml'), Loader=yaml.FullLoader)
        #  node = {}
        #  node['nodeid'] = nodeconf['nodeid']
        #  node['apikey'] = nodeconf['apikey']
        #  node['node'] = args.node
        #  node['user'] = node_user
        #  node['password'] = node_password
        #  cred['nodes'] = [node]
        #except:
        #  print(Fore.YELLOW + 'WARNING: Could not load Node Configuration (pilotnode not configured on target?). Continuing without it.')

        #with open(os.path.join(args.workdir, 'credentials.json') if args.workdir else './credentials.json', 'w') as credfile:
        #  json.dump(cred, credfile)

        # create .pilotfwconfig.json
        config = {}
        config['compiler'] = args.compiler
        config['generated_by'] = "pilot-config v{}".format(pilotconfig_version)
        config['firmware_version'] = firmware_version
        config['config'] = {
          "stop_plc_on_module_error": False,
          "watchdog_timeout": 100,
          "enable_rtc": False
          }
        config['modules'] = []

        for mod in modules:
            if mod['currentfid']:
                module = {}
                module['slot'] = mod['module']
                module['fid'] = mod['currentfid']
                module['fid_nicename'] = mod['currentfid_nicename']
                if 'hid' in mod and 'title' in mod and 'subtitle' in mod and 'description' in mod:
                    module['hid'] = mod['hid']
                    module['title'] = mod['title']
                    module['subtitle'] = mod['subtitle']
                    module['description'] = mod['description']

                devicefile = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'devices', module['fid'] + '.py')
                if os.path.isfile(devicefile):
                    spec = importlib.util.spec_from_file_location("module.name", devicefile)
                    device = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(device)
                    module['config'] = device.default_config()
                config['modules'].append(module)

        if args.node:
            config['nodes'] = [{'name': 'default', 'node': args.node}]

        helper.save_config(args, config)


        project_type = 'default'
        sourcepath = os.path.join(os.path.dirname(os.path.realpath(__file__)),'project', args.compiler.lower(), project_type)

        if args.local:
            # check if a compressed version exists
            source_gz = sourcepath+".tar.gz"
            if (os.path.exists(source_gz)):
                with tarfile.open(source_gz, "r:gz") as tar:
                    tar.extractall(path=targetpath)
                    tar.close()
            elif os.path.exists(sourcepath):
                distutils.dir_util.copy_tree(sourcepath, targetpath)
            else:
                print("Project type '{}' not found".format(project_type))
                exit(1)

        print("Project generated")
        print("""
        Run `pilot fw build` to compile the project
        and `pilot fw program` to program it to the Pilot Mainboard
        """)
        #print( """{}
        #├─ src/
        #│  ├─ program.st    /* IEC 61131-3 code */
        #│  └─ *.c, *.h      /* custom C code compiled into firmware image */
        #├─ .pilotfwconfig.json      /* firmware configuration (memory mapping, module configuration, etc.) */
        #├─ credentials.json /* authentication credentials (sensitive data) */
        #└─ basefw/          /* firmware base code folder */""".format(args.workdir if args.workdir else os.getcwd()))
    except Exception as error:
        print('An error occured creating the project')
        print(error)
        exit(1)

def update(args):
    print("Updating base firmware")
    toplevel = helper.find_fw_toplevel(args)
    if toplevel != '':
        args.workdir = toplevel
        args = helper.get_modules_from_config(args)

        version, modules = download_base_firmware(args)
        helper.update_firmware_version_in_config(args, version)
    else:
        print("Could not find project configuration file '.pilotfwconfig.json', firmware not updated")

def main(args, version, mode):
    if mode == 'init':
        init(args, version)
    elif mode == 'update':
        update(args)



import os
import json
import lazy_import
import importlib.util

#from collections import defaultdict
defaultdict = lazy_import.lazy_callable("collections.defaultdict")
fnmatch = lazy_import.lazy_module("fnmatch")

class Plc():
  targetdevice = { 'flashaddress': 0x08000000, 'ramaddress': 0x20000000, 'bitbandaddress': 0x22000000, 'flash': 512, 'ram': 48 } # STM32F103CBxx
  model = {}
  config = {}
  compiler = None
  helpers = None
  args = {}

  def __init__(self, args, version, model, config, compiler, helpers):
    self.args = args
    self.config = config
    self.model = model
    self.compiler = compiler
    self.helpers = helpers

    self.model['version'] = version

    if 'config' in config:
      self.model['config'] = config['config']

    mem_modules = self.init_memory_mapped_modules()

  def init_memory_mapped_modules(self):
    allmodules = self.config["modules"]
    if 'plugins' in self.config:
      allmodules = allmodules + self.config['plugins']

    # Initialize device data 
    # create device properties
    # spec - hw specs (I/Os, GPIO banks, Interrupts, etc)
    # name - name to be used in code gen
    # absaddress - absolute RAM address in MCU
    # NOTE that 'Slot' in spec is zero based whereas 'slot' is 1 based.
    memmodules = []
    self.config['plc_memory_size'] = 0
    for mod in allmodules:
      dev = {}
      mod['device'] = dev
      try: # will fail if module is not supported by PLC firmware
        dev['hw'] = next(iter(next(iter([v for k, v in self.model.items() if isinstance(v, list) and 
               [x for x in v if k != 'Modules' and 'Slot' in x and 'slot' in mod and x['Slot']+1 == int(mod['slot'])]]))))
      except:
        dev['hw'] = {}
      dev['name'] = "m{}_{}".format(mod['slot'], mod['fid']) # TODO naming for plugins
      dev['slot'] = mod['slot']
      dev['index'] = int(mod['slot']) - 1
      dev['spec'] = self.getDevice(mod)
      if dev['spec'] == None:
        print("Warning: No device found for module firmware {}, this module firmware does not have PLC support".format(mod['fid']))
      else:
        self.config['plc_memory_size'] = self.config['plc_memory_size'] + dev['spec'].size
        dev['spec'].compile()
              
        memmodules.append(mod)

    # memory size in kb
    plcmemsize = int((self.config['plc_memory_size']  + 1024 - 1) / 1024)

    # memory mapped modules
    self.model['memmodules'] = memmodules 
    uniquemodules = []
    for m in memmodules:
      if len(list(filter(lambda x: x['fid'] == m['fid'], uniquemodules))) == 0:
        uniquemodules.append(m)
    self.model['uniquemodules'] = uniquemodules 

    # generate data for linker script
    self.model['memory'] = self.targetdevice
    self.model['ld'] = { 'flashaddress': hex(self.targetdevice['flashaddress']),
                          'plcaddress': hex(self.targetdevice['ramaddress']),
                          'ramaddress': hex(self.targetdevice['ramaddress'] + (plcmemsize * 1024)),
                          'flash': str(self.targetdevice['flash']) + 'k', 
                          'ram': str(int(self.targetdevice['ram']) - plcmemsize) + 'k',
                          'plc': str(int(plcmemsize)) + 'k'}
    
  def getDevice(self, module):
    devicefile = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'devices', module['fid'] + '.py') 
    if not os.path.isfile(devicefile) and 'plugin' in module: # not found in devices, check in plugins folder
      devicefile = os.path.join(self.args.workdir, 'plugins', module['plugin'] + '.py') 
      if (os.path.isfile(devicefile)): # plugin exists, add plugin to model
        if not 'plugins' in self.model:
          self.model['plugins'] = {}
        self.model['plugins']['dcmctrl'] = { 'module': module }

    if os.path.isfile(devicefile):
      spec = importlib.util.spec_from_file_location("module.name", devicefile)
      device = importlib.util.module_from_spec(spec)
      spec.loader.exec_module(device)
      return device.getDevice(self.model, module, self.compiler, self.helpers)
    return None


#!/usr/bin/env python3
import lazy_import

import argparse
import os
import sys
import cryptography

# needed until paramiko is fixed
import warnings
from cryptography import utils

try:
  warnings.simplefilter("ignore", cryptography.utils.CryptographyDeprecationWarning)
except:
  pass
  
bugsnag = lazy_import.lazy_module("bugsnag")

from . import arguments

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),'VERSION'), 'r') as myfile:
    version=myfile.read().replace('\n', '')

def my_except_hook(exectype, value, traceback):
  if exectype == KeyboardInterrupt:
    print("Keyboard Interrupt, cancelling. Bye!")
    exit(3)
  else:
    bugsnag.notify(Exception('Unhandled Exception, type: {}, value: {}, traceback: {}'.format(exectype, value, traceback)))
    sys.__excepthook__(exectype, value, traceback)

sys.excepthook = my_except_hook

def main():
  with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'VERSION'), encoding='utf-8') as f:
    VERSION = f.read().strip()

  bugsnag.configure(
      api_key="2d614bb3561d92fbd3b6f371e39b554a",
      project_root=".",
      app_version=VERSION,
      release_stage='production'
  )

  sys.excepthook = my_except_hook

  # Parent parser for common arguments
  parent_parser = argparse.ArgumentParser(add_help=False)
  parent_parser.add_argument('--hardware', dest='hardware', #choices=[el['name'] for el in targethardwarelist],
                         help='set target hardware')
  parent_parser.add_argument('--server', '-s', default=None, dest='server',
                  help='Alternative URL for the pilot server API to contact')
  parent_parser.add_argument('--node', '-o', default=None, dest='node',
                         help='Node to configure remotely')
  parent_parser.add_argument('--user', '-u', default='pi', dest='user',
                         help='Remote SSH User')
  parent_parser.add_argument('--password', '-p', nargs='?', const='password_was_not_given', dest='password',
                         help='Remote SSH Password')
  parent_parser.add_argument('--sshkey', '-k', nargs='?', default='~/.ssh/id_rsa.pub', dest='sshkey_file',
                         help='public key file')

  argparser = argparse.ArgumentParser(description='Pilot Command-Line Interface')
  argparser.add_argument('-v', '--version', dest='version',action='store_true',
                         help='Get version')

  # Subparsers
  subparsers = argparser.add_subparsers(dest='subparser_name')

  # Setup Subparser
  parser_a = subparsers.add_parser('setup', parents=[parent_parser], help="Configure Pilot firmware")
  arguments.setup_arguments(parser_a)

  parser_r = subparsers.add_parser('reset', parents=[parent_parser], help="Reset Pilot Mainboard")
  arguments.reset_arguments(parser_r)

  parser_m = subparsers.add_parser('module', parents=[parent_parser], help="Get info on modules")
  arguments.module_arguments(parser_m)

  # Firmware subparser
  parser_b = subparsers.add_parser('fw', parents=[parent_parser], help="Init/Build/Program custom firmware")
  parser_b.add_argument('--show-toplevel', dest='show_toplevel', action='store_true',
                        help='show top level folder of firmware project')
  parser_b.add_argument('--show-compilers', dest='show_compilers', action='store_true', help='List all installed compilers for additional firmware (excluding C compiler, this compiler is always installed)')

  fw_subparser = parser_b.add_subparsers(dest='fw_subparser_name')

  parser_d_a = fw_subparser.add_parser('init', parents=[parent_parser], help="Initialize a new firmware project")
  arguments.project_arguments(parser_d_a)

  parser_d_b = fw_subparser.add_parser('update', parents=[parent_parser], help="Update base firmware in project")
  arguments.project_arguments(parser_d_b)

  parser_b_b = fw_subparser.add_parser('build', parents=[parent_parser], help="Compile additional software into firmware (IEC 61131-3 or Rust)")
  arguments.compiler_arguments(parser_b_b)

  parser_b_c = fw_subparser.add_parser('program', parents=[parent_parser], help="Remote program Pilot Mainboard")
  arguments.program_arguments(parser_b_c)

  args = argparser.parse_args()

  if args.version:
    print(VERSION)
  #elif args.modules:
  #  from . import moduleinfo
  #  sys.exit(moduleinfo.main(args))
  elif ('subparser_name' in args):
    if (args.subparser_name == 'setup'):
      from . import pilotsetup
      sys.exit(pilotsetup.main(args))
    elif (args.subparser_name == 'reset'):
      from . import reset
      sys.exit(reset.main(args))
    elif (args.subparser_name == 'module'):
      from .modulehelp  import ModuleHelp
      helper = ModuleHelp()
      sys.exit(helper.help(args))
    elif (args.subparser_name == 'fw'):
      if 'fw_subparser_name' in args and args.fw_subparser_name == 'build':
        from . import build
        sys.exit(build.main(args, VERSION))
      elif 'fw_subparser_name' in args and args.fw_subparser_name == 'program':
        from . import program
        sys.exit(program.main(args))
      elif 'fw_subparser_name' in args and (args.fw_subparser_name == 'init' or args.fw_subparser_name == 'update'):
        from . import project
        sys.exit(project.main(args, VERSION, args.fw_subparser_name))
    else:
      print('No parameters specified, running setup.')
      from . import pilotsetup
      sys.exit(pilotsetup.main(parser_a.parse_args()))

if __name__ == '__main__':
  main()



import lazy_import

import os
import jwt
import time
import datetime
sys = lazy_import.lazy_module("sys")
time = lazy_import.lazy_module("time")
requests = lazy_import.lazy_module("requests")
bugsnag = lazy_import.lazy_module("bugsnag")
os = lazy_import.lazy_module("os")
yaml = lazy_import.lazy_module("yaml")
getpass = lazy_import.lazy_module("getpass")
json = lazy_import.lazy_module("json")
qrcode_terminal = lazy_import.lazy_module("qrcode_terminal")
from uuid import uuid4
Enum = lazy_import.lazy_callable("enum.Enum")
#from enum import Enum

from gql import gql, Client
from gql.transport.requests import RequestsHTTPTransport

from colorama import Fore
from colorama import Style
from colorama import init

from .sbc import Sbc

class RegisterNodeStatus(Enum):
  FAILED = -1
  OK = 0
  CODE_NOT_FOUND = 1
  COULD_NOT_CREATE = 2
  COULD_NOT_UPDATE = 3
  MAC_ALREADY_USED = 4

class PilotServer():
  pilot_server = 'https://gql.pilotnexus.io/v1/graphql'
  oauth_token_url = "https://amescon.eu.auth0.com/oauth/token"
  oauth_device_code_url = "https://amescon.eu.auth0.com/oauth/device/code"

  pilot_dir = '/etc/pilot/'
  nodeconfname = 'pilotnode.yml'
  authfilename = 'auth.json'

  pilot_home_dir = os.path.expanduser('~/.pilot')
  authfile = os.path.join(pilot_home_dir, authfilename)

  client_id = 'hG0Kh6oMY6A2dMUjyjAbTQPTcd8syl58'
  audience = [ "https://api.pilotnexus.io", "https://amescon.eu.auth0.com/userinfo" ]

  keys = None
  tokenset = None
  decoded = None

  terminal = True
  sbc = None

  def __init__(self, sbc: Sbc):
    self.sbc = sbc
    if not os.path.exists(self.pilot_home_dir):
      os.makedirs(self.pilot_home_dir)
    
    try:
      keys = requests.get("https://amescon.eu.auth0.com/.well-known/jwks.json").json()['keys']
      self.keys = [ key for key in keys if key['alg'] == 'RS256' ][0]
    except:
      pass

    

  def loadnodeconf(self):
    nodeconffile = self.pilot_dir + self.nodeconfname
    nodeconf = None
    try:
      nodeconf = yaml.load(self.sbc.getFileContent(nodeconffile), Loader=yaml.FullLoader)
    except:
      nodeconf = None
    return nodeconf
  
  def savenodeconf(self, nodeconf):
    nodeconffile = self.pilot_dir + self.nodeconfname
    nodeconfcontent = yaml.dump(nodeconf, default_flow_style=False)
    return self.sbc.setFileContent(nodeconffile, nodeconfcontent)
  
  def loadnodeauth(self):
    nodeauthfile = self.pilot_dir + self.authfilename
    nodeauth = None
    try:
      nodeauth = json.loads(self.sbc.getFileContent(nodeauthfile))
    except:
      nodeauth = None
    return nodeauth

  def savenodeauth(self, nodeauth):
    nodeauthfile = self.pilot_dir + self.authfilename
    nodeauthcontent = json.dumps(nodeauth)
    return self.sbc.setFileContent(nodeauthfile, nodeauthcontent)

  def authenticate(self):

    # check if there is an node authfile on the host
    nodeauth = self.loadnodeauth()
    if nodeauth != None: #there is one, load
      self.tokenset = nodeauth
      self.decode()
      if self.decoded != None: # token was ok
        return

    payload = {'client_id': self.client_id, 'scope':'openid email offline_access', 'prompt': 'consent' }
    headers = { 'content-type': "application/x-www-form-urlencoded" }

    try:
      response = requests.post(self.oauth_device_code_url, data=payload, headers=headers).json()
    
      print("\n\n")
      print("Open {}{}{} and enter".format(Style.BRIGHT, response['verification_uri'], Style.NORMAL))
      print("\n\n")
      print("""=======>      {}{}{}       <=======""".format(Style.BRIGHT, response['user_code'], Style.NORMAL))
      print("\n\nor scan this code with your Camera app to skip entering the code")
      qrcode_terminal.draw(response['verification_uri_complete'])
      print("note: this code expires in {} minutes".format(response['expires_in'] / 60))

      payload2 = { 'grant_type': "urn:ietf:params:oauth:grant-type:device_code", 
                   'client_id': self.client_id, 'device_code': response['device_code'] }

      done = False
      interval = int(response['interval'])

      while (not done):
        response2 = requests.post(self.oauth_token_url, data=payload2, headers=headers).json()
        if 'error' in response2:
          if response2['error'] == 'authorization_pending':
            time.sleep(interval)
          elif response2['error'] == 'access_denied':
            print(Fore.RED + 'Flow cancelled')
            done = True
          elif response2['error'] == 'expired_token':
            print(Fore.RED + 'Token is expired')
            done = True
          else:
            print(Fore.RED + 'Error {}; {}'.format(response2['error'], response2['error_description']))
            done = True
        elif 'access_token' in response2:
          print(Fore.GREEN + "\n\nYour device was sucessfully authorized.")
          self.tokenset = response2
          self.decode()
          self.savenodeauth(response2)
          done = True

    except Exception as e: 
      print(Fore.RED + 'Error authenticating the device {}'.format(e))
      print(e)

  def refresh(self, decode=True):
    try:
      if self.tokenset != None and 'refresh_token' in self.tokenset:
        headers = { 'content-type': "application/x-www-form-urlencoded" }
        payload = { 'grant_type': "refresh_token", 
                    'client_id': self.client_id,
                    'refresh_token': self.tokenset['refresh_token']
                  }
        response = requests.post(self.oauth_token_url, data=payload, headers=headers).json()

        if 'access_token' in response:
          response['refresh_token'] = self.tokenset['refresh_token']
          self.tokenset = response
          with open(self.authfile, 'w') as authfile:
            json.dump(self.tokenset, authfile)
          if decode:
            self.decode()
          return True
    except:
      print(Fore.RED + 'Error getting refresh token {}'.format(e))
      print(e)

    return False
  
  def decode(self):
    self.decoded = None
    if self.tokenset != None and self.keys != None:
      try:
        pubkey = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(self.keys))
        self.decoded = jwt.decode(self.tokenset['access_token'], pubkey, algorithms='RS256', audience = self.audience) #, audience=self.config.CLIENT_ID) 
      except jwt.ExpiredSignatureError:
        try:
          self.refresh(False) # do not decode, might cause endless loop 
          self.decoded = jwt.decode(self.tokenset['access_token'], pubkey, algorithms='RS256', audience = self.audience) #, audience=self.config.CLIENT_ID) 
        except Exception as e: #hopeless, give up
          print(Fore.RED + 'error decoding token')
          print(e)
      except Exception as e:
        print(Fore.RED + 'error decoding token')
        print(e)
  
  def get_token(self):
    if self.tokenset == None:
      try:
        with open(self.authfile) as authfile:
          self.tokenset = json.load(authfile)
          self.decode()
      except:
        pass

    if self.decoded == None:
      self.authenticate()
    
    if self.decoded == None:
      print(Fore.RED + 'Error, cannot authenticate')
      exit(1)
    
    expires = int(self.decoded['exp'] - time.time())
    
    #print('expires in {} seconds ({} hours)'.format(expires, str(datetime.timedelta(seconds=expires))))

    if expires < 60: # expires in less than 60 seconds
      if not self.refresh():
        print(Fore.RED + 'Cannot refresh authentication token, please authenticate again')
        self.authenticate()
        if self.decoded == None:
          print(Fore.RED + 'Error, cannot authenticate')
          exit(1)

    # here we should have a valid access token
    return 'Bearer {}'.format(self.tokenset['access_token'])

  def query(self, query, variables=None, isPublic=False):
    transport = None
    if isPublic:
      transport = RequestsHTTPTransport( url=self.pilot_server, use_json=True )
    else:
      transport = RequestsHTTPTransport( url=self.pilot_server, use_json=True, headers={ 'Authorization': self.get_token() } )

    client = Client( transport=transport, fetch_schema_from_transport=True )
    return client.execute(query,  variable_values=variables)
    

  def getuserid(self):
    if self.decoded != None and 'https://hasura.io/jwt/claims' in self.decoded and 'x-hasura-user-id' in self.decoded['https://hasura.io/jwt/claims']:
      return int(self.decoded['https://hasura.io/jwt/claims']['x-hasura-user-id'])
    return -1

  def getnode(self):
    nodeid = None

    nodeconf = self.loadnodeconf()
    if nodeconf != None and 'nodeid' in nodeconf:
      nodeid = nodeconf["nodeid"]


    try:
      result = self.query(gql("""query getnode($nodeid: uuid) {
        pilot_node(where: {id: {_eq: $nodeid} }) {
            id
            name
            pilotconfig
            online
          }
      }
      """), {
        "nodeid": nodeid
      })
      if 'pilot_node' in result and isinstance(result['pilot_node'], list) and len(result['pilot_node']) == 1 and 'id' in result['pilot_node'][0]:
        return result['pilot_node'][0]

    except Exception as e:
      print('Error getting node from server')
      print(e)
    return None
  
  def updatenode(self, fwconfig, name='', description=''):
    try: 
      userid = self.getuserid()
      if userid <= 0:
        print(Fore.RED + "Not authenticated, cannot configure node.")
        return

      node = self.getnode()
      nodeid = node['id'] if node !=None else None

      if nodeid == None:
        nodeid = str(uuid4())
        nodeconf = { 'nodeid': nodeid }
        self.savenodeconf(nodeconf)
      if fwconfig == None:
        fwconfig = {}

      query = gql("""
        mutation upsertNode($nodeid: uuid, $userid: Int, $pilotconfig: jsonb, $name: String, $description: String) {
          insert_pilot_node(objects: {id: $nodeid, pilotconfig: $pilotconfig, name: $name, description: $description, userid: $userid}, 
            on_conflict: {constraint: node_pkey, update_columns: [pilotconfig]}) {
            affected_rows
          }
        }
      """)
      result = self.query(query, {
        "nodeid": nodeid,
        "userid": userid,
        "pilotconfig": fwconfig,
        "name": name,
        "description": description
      })

      if 'insert_pilot_node' in result and 'affected_rows' in result['insert_pilot_node'] and result['insert_pilot_node']['affected_rows'] == 1:
        print(Fore.GREEN + "Successfully updated node")

    except Exception as e:
      print(Fore.RED + 'could not update node')
      print(e)
  

  def registernode(self, fwconfig):

    name = input("Enter a name for this node: ")
    description = input("Enter a description for this node: ")

    self.updatenode(fwconfig, name, description)
  


def getDevice(model, module, compiler, helpers):
  return IO16Device(model, module, compiler, helpers)

def toGPIO(this, items):
  return 'GPIO' + chr(items+65)

def default_config():
  return { 'direction':  {'0-3': 'in', '4-7': 'in', '8-11': 'in', '12-15': 'in'} }

class IO16Device():
  size = 2
  include = ['io16.h'] 
  init_source = ''
  dev_to_mem_source = ''
  mem_to_dev_source = ''
  mem_doc = { "read": [], "write": [] }

  decl = {
    'c': { 'name': 'io16_t', 'decl': 'typedef uint16_t io16_t;' },
    'rust': { 'name': 'u16', 'decl': '' }
  }

  module = None
  model = None
  compiler = None
  helpers = None 
  
  directions = {'0-3': 'in', '4-7': 'in', '8-11': 'in', '12-15': 'in'}

  def __init__(self, model, module, compiler, helpers):
    self.size = 2
    self.type = 'uint16_t'
    self.length = 1
    self.module = module
    self.helpers = {'gpio': toGPIO, **helpers}
    self.compiler = compiler
    self.model = model
  

  def direction_to_enum(self, dir):
    if dir == 'out':
      return 'pilot_io16_direction_output'
    else:
      return 'pilot_io16_direction_input'

  def compile(self):
    if "config" not in self.module:
      self.module['config'] = { 'direction': directions }

    init_str = "  // initialization for device {{device.name}}\n  int32_t status = 0;\n"
    dev_to_mem_str = "// source for device {{device.name}}\n  int16_t {{device.name}}_value;\n"
    mem_to_dev_str = "// source for device {{device.name}}\n  int16_t {{device.name}}_value;\n"
    # iterate nibbles
    nibbles_read = [True, True, True, True]
    for nibble in self.module['config']['direction']:
      if nibble not in self.directions:
        print("directions in IO16 module {} misconfigured, '{}' is not valid".format(self.module['slot'], nibble))
        exit(1)
      dir = self.module['config']['direction'][nibble]
      if dir != 'in' and dir != 'out':
        print("direction needs to be 'in' or 'out' in '{}'".format(nibble))
        exit(1)
      offset = int(nibble.split('-')[0])
      offset_to = int(nibble.split('-')[1])

      nibbles_read[int(offset / 4)] = dir == 'in'
      init_str = init_str + "  status |= pilot_io16_{{device.index}}_set_direction(pilot_io16_block_" + str(offset) + "_to_" + str(offset_to) + ", " + self.direction_to_enum(dir) + ");\n"

      for i in range(4):
        self.mem_doc['read' if dir == 'in' else 'write'].append({ "name": "{}{}".format(dir[0], offset+i), "desc": "digital {} {}".format('input' if dir == 'in' else 'output', offset+i), "byte": int((offset+i) / 8), "bit": (offset+i) % 8, "length": 1, "signed": False, "datatype": "bool"})

    init_str = init_str + "  return status;"

    readmask = "0x" + "".join(map(lambda x: 'F' if x else '0', reversed(nibbles_read)))
    writemask = "0x" + "".join(map(lambda x: '0' if x else 'F', reversed(nibbles_read)))
    inv_readmask = "0x" + "".join(map(lambda x: '0' if x else 'F', reversed(nibbles_read)))
    inv_writemask = "0x" + "".join(map(lambda x: 'F' if x else '0', reversed(nibbles_read)))
    
    if readmask != '0x0000':
      dev_to_mem_str = dev_to_mem_str + """  get_module_info()->m{{device.index}}_status |= pilot_io16_{{device.index}}_read_register(pilot_io16_register_input_register_A, 2, (uint8_t *)&{{device.name}}_value);
  plc_mem_devices.m{{device.slot}} &= """ + inv_readmask + """; 
  plc_mem_devices.m{{device.slot}} |= {{device.name}}_value & """ + readmask + ";\n"

    if writemask != '0x0000':
      mem_to_dev_str = mem_to_dev_str + "  {{device.name}}_value = plc_mem_devices.m{{device.slot}} & " + writemask + ";\n  " + "get_module_info()->m{{device.index}}_status |= pilot_io16_{{device.index}}_write_register(pilot_io16_register_output_register_A, 2, (uint8_t *)&{{device.name}}_value);"

    # generate source 
    init_template = self.compiler.compile(init_str)
    dev_to_mem_template = self.compiler.compile(dev_to_mem_str)
    mem_to_dev_template = self.compiler.compile(mem_to_dev_str)

    self.init_source = init_template(self.module, self.helpers)
    self.dev_to_mem_source = dev_to_mem_template(self.module, self.helpers)
    self.mem_to_dev_source = mem_to_dev_template(self.module, self.helpers)


import paramiko
#import base64

client = paramiko.SSHClient()
client.load_system_node_keys()
client.set_missing_node_key_policy(paramiko.WarningPolicy())

client.connect('192.168.101.85', username='pi', password='raspberry')
#stdin, stdout, stderr = client.exec_command('cat /proc/pilot/module3/eeprom/fid')
stdin, stdout, stderr = client.exec_command('uname -a')
output = ''.join(stdout)
print(output)

chan = client.get_transport().open_session()
chan.exec_command('uuname -a')
print ("exit status: {}".format(chan.recv_exit_status()))

client.close()



from sys import modules
from halo import Halo
import lazy_import

sys = lazy_import.lazy_module("sys")
time = lazy_import.lazy_module("time")
re = lazy_import.lazy_module("re")
os = lazy_import.lazy_module("os")
scp = lazy_import.lazy_module("scp")
tarfile = lazy_import.lazy_module("tarfile")
itertools = lazy_import.lazy_module("itertools")
requests = lazy_import.lazy_module("requests")
bugsnag = lazy_import.lazy_module("bugsnag")

grpc = lazy_import.lazy_module("grpc")
#import grpc

import traceback

from pathlib import Path

from .pilotserver import PilotServer
from .sbc import Sbc

from .grpc_gen.pilotbuild_pb2 import BinaryType, BuildRequest, BuildStatus, BuildStep, TargetMcu, Module
from .grpc_gen.pilotbuild_pb2_grpc import PilotBuildStub

from shutil import copyfile
from gql import gql

from colorama import Fore
from colorama import Style
from colorama import init

from enum import Enum


class PilotDriver():
    MODULE_COUNT = 4
    retry_count = 2
    emptystr = '-'
    build_server = 'build.pilotnexus.io:50051'
    pilot_driver_root = '/proc/pilot'
    tmp_dir = '/tmp/pilot'
    kernmodule_list = [
        'pilot', 'pilot_plc', 'pilot_tty', 'pilot_rtc', 'pilot_fpga',
        'pilot_io', 'pilot_slcd'
    ]

    binpath = ''

    eeproms = {}
    modules = {}

    ps = None
    sbc = None

    def __init__(self, pilotserver: PilotServer, sbc: Sbc):
        self.ps = pilotserver
        self.sbc = sbc

        if sbc is not None:
            self.binpath = '{}/bin/{}'.format(
                os.path.join(os.path.abspath(os.path.dirname(__file__))),
                self.sbc.target['architecture'])

    def get_modules(self):
        memregs = ['uid', 'hid', 'fid']
        strmlist = list(
            filter(
                None,
                self.sbc.cmd(
                    'find /proc/pilot/module* -maxdepth 0 -printf "%f\\n"').
                split('\n')))
        matches = filter(None,
                         map(lambda x: re.match(r'module(\d+)', x), strmlist))
        modlist = {
            int(x.group(1)): {
                'uid': '',
                'hid': '',
                'fid': ''
            }
            for x in matches
        }
        for mod in modlist:
            for memreg in memregs:
                retry = self.retry_count
                errcount = 0
                while retry > 0:
                    retry = retry - 1
                    try:
                        regfile = self.sbc.cmd(
                            'cat {}/module{}/eeprom/{}'.format(
                                self.pilot_driver_root, mod, memreg), True)
                        modlist[mod][memreg] = ''.join(
                            char for char in regfile
                            if str.isprintable(char)).strip()

                        if memreg == 'fid':  # additional plausibility checks
                            if len(
                                    list(
                                        filter(lambda x: ord(x) > 255,
                                               regfile))) > 0:
                                modlist[mod][memreg] = ''
                            else:
                                break
                        else:
                            break
                    #except:
                    except Exception as e:
                        modlist[mod][memreg] = ''
                        errcount = errcount + 1
                        if errcount >= self.retry_count:
                            return modlist, False
                        #dbg('could not read {} of module {}'.format(memreg, mod))
        return modlist, True

    def set_module_fid(self, number, fid):
        self.sbc.setFileContent(
            self.pilot_driver_root + '/module{}/eeprom/fid'.format(number),
            fid + '       ')

    def getmodules(self, mod):
        self.eeproms = mod
        try:
            variables = {
                'fids': [
                    '{}'.format(value['fid']) for key, value in mod.items()
                    if value['fid'] != ''
                ],
                'hids': [
                    '{}'.format(value['hid']) for key, value in mod.items()
                    if value['hid'] != ''
                ]
            }

            obj = self.ps.query(
                gql("""
      query get_fids($fids: [bpchar!], $hids: [bpchar!]) {
        pilot_fid(where: {fid: {_in: $fids} }) {
          fid
          name
          usage
        }
        pilot_hid(where: {hid: {_in: $hids} }) {
          hid
          title
          subtitle
          description
          pinout
          hid2fids {
            fid {
              fid
              name
            }
            isdefault
          }
        }
      }
      """), variables, True)

            self.modules = [{
                'module':
                key,
                'currentfid': value['fid'],
                'hid': value['hid'],
                'currentfid_nicename':
                next(
                    iter(y['name'] for y in filter(
                        lambda x: x['fid'].strip() == value['fid'].strip(),
                        obj['pilot_fid'])), value['fid'].strip()),
                'usage':
                next(
                    iter(y['usage'] for y in filter(
                        lambda x: x['fid'].strip() == value['fid'].strip(),
                        obj['pilot_fid'])), value['fid'].strip()),
                'description':
                next(
                    iter(y['description'] for y in filter(
                        lambda x: x['hid'].strip() == value['hid'].strip(),
                        obj['pilot_hid'])), value['hid'].strip()),            
                'pinout':
                next(
                    iter(y['pinout'] for y in filter(
                        lambda x: x['hid'].strip() == value['hid'].strip(),
                        obj['pilot_hid'])), value['hid'].strip()),            
                'fids':
                next(
                    iter([{
                        'isdefault': x['isdefault'],
                        'fid': x['fid']['fid'],
                        'name': x['fid']['name']
                    } for x in y['hid2fids']] for y in filter(
                        lambda x: x['hid'].strip() == value['hid'].strip(),
                        obj['pilot_hid'])), []),
            } for key, value in mod.items()]

            return self.modules, True
        except:
            e = sys.exc_info()[0]
            print(e)
        return {}, False

    def load_pilot_defs(self):
        success = False
        try:
            eeproms, success = self.get_modules()
            return self.getmodules(eeproms)
        except:
            return {}, False

    def driver_loaded(self):
        return self.sbc.cmd_retcode('test -e {}'.format(
            self.pilot_driver_root)) == 0

    def getModuleEeprom(self, module, memregion):
        try:
            return self.sbc.getFileContent('{}/module{}/eeprom/{}'.format(
                self.pilot_driver_root, module, memregion))
        except:
            return ''

    def tryrun(self, text, retries, cmd):
        print(text + '...', end='')
        sys.stdout.flush()
        while retries > 0:
            retries = retries - 1
            try:
                if self.sbc.cmd_retcode(cmd) == 0:
                    print(Fore.GREEN + 'done')
                    return 0
            except:
                print('.', end='')
                sys.stdout.flush()
        print(Fore.RED + 'failed')
        return 1

    def reset_pilot(self, wait_bootmsg=False):
        try:
            if not wait_bootmsg:
                print('Resetting MCU...', end='')
            sys.stdout.flush()
            reset_pin = self.sbc.target['reset_pin']['number']
            boot_pin = self.sbc.target['boot_pin']['number']

            # set boot pin to 0 to avoid booting into bootloader
            self.sbc.cmd(
                'sudo sh -c \'[ ! -f /sys/class/gpio/gpio{0}/value ] && echo "{0}" > /sys/class/gpio/export\''
                .format(boot_pin))
            self.sbc.cmd(
                'sudo sh -c \'echo "out" > /sys/class/gpio/gpio{}/direction\''.
                format(boot_pin))
            self.sbc.cmd(
                'sudo sh -c \'echo -n "0" > /sys/class/gpio/gpio{}/value\''.
                format(boot_pin))

            self.sbc.cmd(
                'sudo sh -c \'[ ! -f /sys/class/gpio/gpio{0}/value ] && echo "{0}" > /sys/class/gpio/export\''
                .format(reset_pin))
            self.sbc.cmd(
                'sudo sh -c \'echo "out" > /sys/class/gpio/gpio{}/direction\''.
                format(reset_pin))
            self.sbc.cmd(
                'sudo sh -c \'echo -n "1" > /sys/class/gpio/gpio{}/value\''.
                format(reset_pin))
            time.sleep(2)
            if wait_bootmsg:
                missing_commands = ''
                if self.sbc.cmd_retcode('command -v stty') != 0:
                    missing_commands = missing_commands + 'stty '
                if self.sbc.cmd_retcode('command -v timeout') != 0:
                    missing_commands = missing_commands + 'timeout '

                if missing_commands == '':
                    return self.sbc.cmd(
                        "sudo sh -c 'stty -F {0} 921600;sleep 0.1;timeout 2 cat {0} & sleep 0.5; echo 0 > /sys/class/gpio/gpio{1}/value;wait'"
                        .format(self.sbc.target['tty'], reset_pin))
                else:
                    return self.sbc.cmd(
                        "/sys/class/gpio/gpio{0}/value".format(reset_pin))
            else:
                self.sbc.cmd(
                    'sudo sh -c \'echo -n "0" > /sys/class/gpio/gpio{}/value\''
                    .format(reset_pin))
                print(Fore.GREEN + 'done')
        except:
            print(Fore.RED + 'failed')
        return ''

    def install_driver(self):
        try:
            match = re.match(self.sbc.target['kernelversionre'],
                             self.sbc.cmd('uname -a'))
            if match:
                packagename = "pilot-{}{}".format(
                    match.group('version'),
                    match.group('buildnum')
                    if 'buildnum' in match.groupdict() else '')
                print('trying to install package '
                      '{}'
                      ''.format(packagename))
                if self.sbc.cmd_retcode(
                        """sudo sh -c 'echo "{}" > /etc/apt/sources.list.d/amescon.list'"""
                        .format(self.sbc.target['apt_source'])) != 0:
                    print(
                        'Could not add source to /etc/apt/sources.list.d/amescon.list'
                    )
                    return 1
                if self.sbc.cmd_retcode(
                        """sudo sh -c 'wget -qO - http://archive.amescon.com/amescon.asc | sudo apt-key add -'"""
                ) != 0:
                    print('Could not get signing keys from amescon keyserver')
                    return 1
                self.sbc.cmd_retcode('sudo apt-get update')
                self.sbc.cmd_retcode("sudo apt-get remove '^pilot-.*' -y")
                self.sbc.cmd('sudo apt-get install -y {}'.format(packagename),
                             True)
            else:
                print('Could not detect your linux version')
                return 1
        except Exception as e:
            # bugsnag.notify(e, user={
            #     "username": self.ps.pilotcfg['username']})
            return 1
        return 0

    def get_kernel_info(self):
        try:
            fwhash = self.sbc.cmd(
                "echo $(/bin/zgrep '* firmware as of' /usr/share/doc/raspberrypi-bootloader/changelog.Debian.gz | head -1 | awk '{ print $5 }') && /usr/bin/wget https://raw.github.com/raspberrypi/firmware/$FIRMWARE_HASH/extra/git_hash -O - 2> NUL",
                True)
            if self.sbc.cmd_retcode('sudo modprobe configs') == 0:
                print(
                    'If you want us to build a kernel for you, send us the following firmware hash AND the file on your Raspberry Pi located in: /proc/config.gz'
                )
                print('Firmware Hash: {}'.format(fwhash))
                return 0
            else:
                print(
                    'Could not load configs. Check if the target Hardware is a Raspberry Pi and your user is sudoer.'
                )
        except:
            print(
                "Could not get your Firmware Hash. Check if the target Hardware is a Raspberry Pi and it is online."
            )
        return 1

    def check_driver(self):
        if self.sbc.cmd_retcode('ls ' + self.pilot_driver_root) != 0:
            print('Pilot driver is not loaded. Trying to load')
            if self.reload_drivers(False):
                print('Drivers loaded')
            else:
                if self.install_driver() == 0:
                    print('Pilot driver installed.')
                    return 1
                else:
                    print('Could not install the pilot driver, most likely there is no pre-compiled driver for your kernel.')
                    ch = input("Do you want to try to build them locally? (needs a couple of minutes and plenty of disk space, be patient) [y/n]: ")
                    if (ch == 'y' or ch == 'yes'):
                        return self.try_build_drivers()
                    #self.get_kernel_info()
                    return -1
        return 0
    
    def try_build_drivers(self):
        print('Building Pilot Kernel Driver...', end='')
        sys.stdout.flush()
        ret = self.sbc.cmd_retcode("sudo apt-get install -y python2 git build-essential")
        if ret == 0:
            ret = self.sbc.cmd_retcode("git clone https://github.com/pilotnexus/pilotdriver.git ~/pilotdriver && cd ~/pilotdriver && make prepare && make && make package && sudo make install")
            if ret == 0:
                print(Fore.GREEN + 'done')
                return 1 # success, we built the drivers, returning 1 to indicate a reboot
        print(Fore.RED + 'failed')
        print("Could not install build tools")
        return ret 

    def reload_drivers(self, verbose=True, pdstopped=False):
        ok = True

        pnstopped = self.sbc.stop_service("pilotnode", verbose)
        if not pdstopped:
            pdstopped = self.sbc.stop_service("pilotd", verbose)

        if verbose:
            print('reloading drivers...', end='')
        sys.stdout.flush()
        for module in self.kernmodule_list[::-1]:
            try:
                self.sbc.cmd_retcode("sudo modprobe -r {}".format(module))
            except:
                pass

        for module in self.kernmodule_list:
            try:
                if self.sbc.cmd_retcode(
                        "sudo modprobe {}".format(module)) != 0:
                    ok = False
            except:
                ok = False

        if verbose:
            if ok:
                print(Fore.GREEN + 'done')
            else:
                print(Fore.RED + 'failed')

        if pdstopped:
            self.sbc.start_service("pilotd")

        if pnstopped:
            self.sbc.start_service("pilotnode")

        return ok

    def build(self, version):
        files = dict()
        try:
            if not os.path.exists(self.tmp_dir):
                os.makedirs(self.tmp_dir)
            if version is not None:
                print("Building Version {}".format(version))
            spinner = Halo(text='Connecting', spinner='dots')
            spinner.start()
            channel = grpc.insecure_channel(self.build_server)
            stub = PilotBuildStub(channel)
            modules = [
                Module(number=key, fid=value['fid'], hid=value['hid'], uid=value['uid'])
                for key, value in self.eeproms.items() if value['fid'] != ''
            ]
            request = BuildRequest(modules=modules,
                                   target=TargetMcu.STM32F103RC, version=version)
            for buildStatus in stub.Build(request=request):
                if (buildStatus.HasField('step')):
                    spinner.stop()
                    text = re.sub("([A-Z])", " \g<0>",
                                  str(BuildStep.Name(buildStatus.step)))
                    spinner = Halo(text=text, spinner='dots')
                    spinner.start()
                    #print(BuildStep.Name(buildStatus.step))
                elif (buildStatus.HasField('result')):
                    spinner.stop()
                    spinner = Halo(
                        text='Downloading ' +
                        str(BinaryType.Name(buildStatus.result.bintype)),
                        spinner='dots')
                    spinner.start()
                    filename = ''
                    extract = False
                    extractfname = ''
                    extractdir = ''
                    if (buildStatus.result.bintype == BinaryType.MCUFirmware):
                        filename = "stm.gz"
                        if buildStatus.result.version:
                            version = buildStatus.result.version
                        extract = True
                        extractdir = self.tmp_dir
                        extractfname = 'stm.bin'
                    elif (buildStatus.result.bintype == BinaryType.MCUSource):
                        filename = "stm_src.gz"
                    elif (buildStatus.result.bintype ==
                          BinaryType.FPGABitstream):
                        filename = "fpga.gz"
                        extract = True
                        extractdir = self.tmp_dir
                        extractfname = 'cpld.jam'
                    elif (buildStatus.result.bintype == BinaryType.FPGASource):
                        filename = "fpga_src.gz"

                    fname = '{}/{}'.format(self.tmp_dir, filename)
                    with open(fname, "wb") as file:
                        file.write(buildStatus.result.binary)

                    if extract:
                        with tarfile.open(fname, "r:gz") as tar:
                            tar.extractall(path=extractdir)
                            tar.close()
                            if extractfname:
                                files[buildStatus.result.
                                      bintype] = os.path.join(
                                          extractdir, extractfname)
                    else:
                        files[buildStatus.result.bintype] = fname

            spinner.stop()
            print("Done.")
            return version, files

        except Exception as error:
            print(Fore.RED + str(error))
            bugsnag.notify(Exception(error))
    
    def get_firmware_source(self, extractDir, version):
        try:
            version, files = self.build(version)
            mcu_src_expanded = False
            fpga_src_expanded = False
            if BinaryType.FPGASource in files:
                with tarfile.open(files[BinaryType.FPGASource], "r:gz") as tar:
                    tar.extractall(path=extractDir)
                    tar.close()
                    fpga_src_expanded = True
            if BinaryType.MCUSource in files:
                with tarfile.open(files[BinaryType.MCUSource], "r:gz") as tar:
                    tar.extractall(path=extractDir)
                    tar.close()
                    mcu_src_expanded = True
            return version, fpga_src_expanded and mcu_src_expanded
        except Exception as error:
            print(Fore.RED + str(error))
            bugsnag.notify(Exception(error))
        return None, False

    def program_cpld(self, binfile, erase=False):
        cmd = 'sudo chmod +x {0}/jamplayer; sudo {0}/jamplayer -a{1} -g{2},{3},{4},{5} {6}'.format(
            self.binpath, 'erase' if erase else 'program',
            self.sbc.target['tdi_pin']['number'],
            self.sbc.target['tms_pin']['number'],
            self.sbc.target['tdo_pin']['number'],
            self.sbc.target['tck_pin']['number'], binfile)
        return self.tryrun('erasing CPLD' if erase else 'programming CPLD', 3,
                           cmd)

    def serial_used(self):
        res = self.sbc.cmd('sudo fuser -v {0}'.format(self.sbc.target['tty']))
        if len(res.strip()) > 0:
            return self.tryrun(
                'killing process accessing tty', 2,
                'sudo fuser -k {0}; sleep 1'.format(self.sbc.target['tty']))
        return True

    def program_mcu( self, binfile):  #use 115200, 57600, 38400 baud rates sequentially
        return self.tryrun(
            'programming MCU', 4,
            'sudo chmod +x {0}/stm32flash; sudo {0}/stm32flash -w {1} -b 115200 -g 0 -x {2} -z {3} {4}'
            .format(self.binpath, binfile, self.sbc.target['reset_pin']['number'],
                    self.sbc.target['boot_pin']['number'],
                    self.sbc.target['tty']))

    def program(self,
                files,
                program_cpld=True,
                program_mcu=True,
                bootmsg=False,
                reload_driver=True):
        res = 0
        if self.sbc.remote_client:
            self.sbc.cmd_retcode('mkdir -p {}'.format(self.tmp_dir))
            if self.sbc.cmd_retcode('sudo chown -R $USER {}'.format(
                    self.tmp_dir)) == 0:
                with scp.SCPClient(
                        self.sbc.remote_client.get_transport()) as scp_client:
                    scp_client.put(self.binpath + '/jamplayer',
                                   remote_path=self.tmp_dir)
                    scp_client.put(self.binpath + '/stm32flash',
                                   remote_path=self.tmp_dir)
                    scp_client.put(self.binpath + '/stm32flash',
                                   remote_path=self.tmp_dir)
                    self.binpath = self.tmp_dir # set binpath to tmp_dir now that executables are copied there

                    # FPGA Bitstream
                    if BinaryType.FPGABitstream in files:
                        scp_client.put(files[BinaryType.FPGABitstream],
                                       remote_path=os.path.join(
                                           self.tmp_dir, 'cpld.jam'))
                    # MCU Firmware
                    if BinaryType.MCUFirmware in files:
                        scp_client.put(files[BinaryType.MCUFirmware],
                                       remote_path=os.path.join(
                                           self.tmp_dir, 'stm.bin'))

                    # Variable file
                    if BinaryType.Variables in files:
                        scp_client.put(files[BinaryType.Variables],
                                       remote_path=os.path.join(
                                           self.tmp_dir, 'variables'))
                    
                    # Docs file
                    if BinaryType.Docs in files:
                        scp_client.put(files[BinaryType.Docs],
                                       remote_path=os.path.join(
                                           self.tmp_dir, 'fwconfig.json'))
            else:
                print('Error setting permissions to folder {}'.format(
                    self.tmp_dir))
        else:
            src = ""
            target = ""
            if BinaryType.FPGABitstream in files:
                src = files[BinaryType.FPGABitstream]
                target = os.path.join(self.tmp_dir, 'cpld.jam')
            if BinaryType.MCUFirmware in files:
                src = files[BinaryType.MCUFirmware]
                target = os.path.join(self.tmp_dir, 'stm.bin')
            if BinaryType.Variables in files:
                src = files[BinaryType.Variables]
                target = os.path.join(self.tmp_dir, 'variables')
            if BinaryType.Docs in files:
                src = files[BinaryType.Docs]
                target = os.path.join(self.tmp_dir, 'fwconfig.json')

            if src != target:
                copyfile(src, target)

        if program_cpld and res == 0:
            res = self.program_cpld(
                Path(self.tmp_dir).joinpath('cpld.jam').as_posix(), True)

        if program_mcu and res == 0:
            pdstopped = self.sbc.stop_service("pilotd", True)
            self.serial_used()
            res = self.program_mcu(
                Path(self.tmp_dir).joinpath('stm.bin').as_posix())

        if program_cpld and res == 0:
            res = self.program_cpld(
                Path(self.tmp_dir).joinpath('cpld.jam').as_posix())

        print(self.reset_pilot(bootmsg))
        if res == 0 and BinaryType.Variables in files:
            res = self.tryrun(
                'setting PLC variables', 4,
                'sudo cp {}/variables /proc/pilot/plc/varconfig'.format(
                    self.tmp_dir))
            self.tryrun(
                'setting PLC variables permanently', 4,
                'sudo mkdir -p /etc/pilot; sudo cp {}/variables /etc/pilot/variables'
                .format(self.tmp_dir))

        if res == 0 and BinaryType.Docs in files:
            res = self.tryrun(
                'saving firmware configuration file', 4,
                'sudo mkdir -p /etc/pilot; sudo cp {}/fwconfig.json /etc/pilot/fwconfig.json'
                .format(self.tmp_dir))

        if reload_driver:
            self.reload_drivers(True, pdstopped)

        return res, pdstopped

    def get_help(self):
        try:
            query = u"""
      {{
        modulehelp(modules: [{}]) {{
          number
          help
          examples {{title  example}}
        }}
      }}
      """.format(','.join([
                '{{number: {}, fid: "{}"}}'.format(key, value['fid'])
                for key, value in self.eeproms.items() if value['fid'] != ''
            ]))
            ret, obj = self.ps.query_graphql(query)
            if ret == 200 and obj['data'] and obj['data']['modulehelp']:
                return obj['data']['modulehelp']
        except:
            e = sys.exc_info()[0]
            print('Could not contact Pilot Nexus API to get help data')
            bugsnag.notify(e)
        return None



from asyncore import file_dispatcher
from pathlib import Path
from .pilotdriver import PilotDriver
from .pilotserver import PilotServer
from .sbc import Sbc

import re

import lazy_import
Compiler = lazy_import.lazy_callable("pybars.Compiler")

from rich.console import Console
from rich.markdown import Markdown
from rich import print as rich_print
from rich.table import Table, Column

from colorama import Fore
from colorama import Style
from colorama import init

init(autoreset=True)  # colorama color autoreset

# finds the largest string in an array
def largest_str(arr):
    # Initialize maximum element
    n = len(arr)
    max = len(arr[0])
    # Traverse array elements from second
    # and compare every element with 
    # current max
    for i in range(1, n):
        cur = len(arr[i])
        if cur > max:
            max = cur
    return max

class ModuleHelp():
    sbc = None
    module = None
    
    def gpiochip_base(self, _, offset):
        p = re.compile('^gpiochip([0-9]+)$', re.MULTILINE)
        l = re.compile('^pilot.*?_([0-9]+)$')
        matches = p.findall(self.sbc.cmd('ls /sys/class/gpio'))
        for gc_str in matches:
            gpio_start = int(gc_str)
            if gpio_start >= 60 and gpio_start < 500: #min gpio number of pilot GPIOs
                label = self.sbc.cmd('cat /sys/class/gpio/gpiochip{}/label'.format(gpio_start))
                num_match = l.match(label)
                if num_match and int(num_match.group(1)) == self.number: 
                    if offset != None:
                        gpio_start += offset
                    return gpio_start
        return -1
    
    def tty(self, _, block, offset=0):
        return "/dev/ttyP{}".format((self.number)*2+offset)


    def help(self, args):
        with Sbc(args) as sbc:
            self.sbc = sbc
            pilotserver = PilotServer(sbc)
            if 'server' in args and args.server != None:
                pilotserver.pilot_server = args.server
            
            #PilotDriver
            pilotdriver = PilotDriver(pilotserver, sbc)
    
            modules, success = pilotdriver.load_pilot_defs()
            if success:
                for mod in modules:
                    if mod['module'] == args.module:
                        self.number = args.module-1
                        self.printhelp(args, mod)
    
    def loop(self, _, block, start, to, incr):
        accum = '' 
        for i in range(start, to, incr):
            accum += "".join(block['fn'](i))
        return accum
    
    def printformatted(self, line):
        splitchars=["#", "//"]
        splitchar=""
        for sp in splitchars:
            splitchar = sp
            parts=line.split(splitchar, 1)
            if (len(parts) > 1):
                break
        if len(parts) > 1:
            print(Style.RESET_ALL + parts[0] + Fore.GREEN + splitchar + parts[1])
        else:
            print(parts[0])

    def printhelp(self, args, module):
            helpers = {'gpio_base': self.gpiochip_base, 'for': self.loop }

            print("")
            print(Fore.GREEN + 'Module [{}]: {}'.format(args.module, module['currentfid_nicename']))
            print("")
    
            if module != None:
                compiler = Compiler()

                if args.pinout:
                    if 'pinout' in module and 'default' in module['pinout'] and 'pins' in module['pinout']['default']:
                        pins = module['pinout']['default']['pins'] 
                        width = largest_str([item for sublist in pins for item in sublist]) + 2 # add a space on each side of the string

                        grid = Table(show_header=False, expand=False)
                        grid.add_column(justify="center")
                        for row in pins:
                            table = Table(show_header=False, expand=False)
                            i = 0
                            for col in row:
                                # color special pins
                                if col == "GND":
                                    row[i] = "[green]" + col
                                elif col == "VCC" or col == "24V" or col == "5V":
                                    row[i] = "[red]" + col
                                table.add_column(justify="center", width=width)
                                i = i + 1
                            table.add_row(*row)
                            grid.add_row(table)
                        rich_print(grid)
                    else:
                        print("No pinout available for this module")
                elif args.usage:
                    if 'usage' in module and module['usage'] != None:
                        template = compiler.compile(module['usage'])
                        text = template({
                            'module': module['module']-1, 
                            'slot': module['module'], 
                            'tty': "/dev/ttyP{}".format((module['module']-1)*2),
                            'tty1': "/dev/ttyP{}".format((module['module']-1)*2),
                            'tty2': "/dev/ttyP{}".format((module['module']-1)*2+1),
                            }, helpers)
                        console = Console()
                        md = Markdown(text)
                        console.print(md)
                    else:
                        print("No usage description available for this module")
                else:
                    if 'description' in module and module['description'] != None:
                        console = Console()
                        md = Markdown(module['description'])
                        console.print(md)
                    else:
                        print("No description available for this module")

                    print("")
                    print("Run with --pinout to see pinout")
                    print("Run with --usage to get usage examples")
            else:
                print("Module not found")






import os
import lazy_import

from bugsnag.handlers import BugsnagHandler
from colorama import init

logging = lazy_import.lazy_module("logging")
# class imports
from .pilotdriver import PilotDriver
from .pilotserver import PilotServer
from .sbc import Sbc

############### INIT ###################

init(autoreset=True)  # colorama color autoreset

DEBUG = False
EXECVP_ENABLED = False

def main(args):

    logger = logging.getLogger()
    handler = BugsnagHandler()
    # send only ERROR-level logs and above
    handler.setLevel(logging.ERROR)
    logger.addHandler(handler)

    logging.getLogger("paramiko").setLevel(logging.ERROR)

    trywritedefaultfirmware = False
    result = 0
    try:
        with Sbc(args) as sbc:
            # PilotServer
            pilotserver = PilotServer(sbc)
            if args.server != None:
                pilotserver.pilot_server = args.server

            #PilotDriver
            pilotdriver = PilotDriver(pilotserver, sbc)

            if not args.node and os.getuid() != 0:
                print('Please run with sudo permissions.')
                return 2

            if sbc.need_sudo_pw():
                print(
                    'we need sudo on remote machine (without interactive authentication)'
                )
                return 2

            print('Resetting Pilot Mainboard')
            print(pilotdriver.reset_pilot(args.wait_bootmsg))
            return 0
    except:
        print("Error resetting Pilot Mainboard")
        return 1


import lazy_import
from colorama import Fore
os = lazy_import.lazy_module("os")
paramiko = lazy_import.lazy_module("paramiko")
logging = lazy_import.lazy_module("logging")
subprocess = lazy_import.lazy_module("subprocess")
fnmatch = lazy_import.lazy_module("fnmatch")
json = lazy_import.lazy_module("json")
import getpass
class Sbc():
  logging.getLogger("paramiko").setLevel(logging.ERROR)

  remote_client = None
  args = None
  targethardwarelist = {}
  target = {}

  def __init__(self, args):
    self.args = args

    # load target hardware list
    for root, _dirnames, filenames in os.walk(os.path.dirname(os.path.realpath(__file__))):
      for filename in fnmatch.filter(filenames, 'targethardware.json'):
        with open(os.path.join(root, filename)) as obj:
          self.targethardwarelist = json.load(obj)['targethardware']

  def __enter__(self):
    usecredentials = False
    usekeyfile = False
    connected = False

    if 'node' in self.args and self.args.node != None:
      if self.args.password != None: # use password was specified
        if self.args.password == 'password_was_not_given':
          self.args.password = getpass.getpass(prompt='Enter SSH password: ', stream=None) 
        usecredentials = True
      else: # no password parameter, usekeyfile
        self.args.sshkey_file = os.path.expanduser(self.args.sshkey_file)
        if os.path.exists(self.args.sshkey_file):
          usekeyfile = True
        else:
          print(Fore.RED + 'error')
          print("Password parameter ('-p' or '--password') not specified, and ssh key file '{}' does not exist.\nPlease specify a password or a valid ssh key file.".format(self.args.sshkey_file))
          exit(1)
      try:
        if usekeyfile:
          self.connect_with_key(self.args.user, self.args.sshkey_file)
          connected = True
        elif usecredentials:
          self.connect(self.args.user, self.args.password)
          connected = True
        else:
          print(Fore.YELLOW + 'failed')
          print("You have neither specified a password nor a valid ssh key file (the default '{}' does not exist either, yes we tried that too).\nPlease specify a password or a valid ssh key file.".format(self.args.sshkey_file))
          exit(1)
      except:
        print(Fore.YELLOW + 'failed')
    else: # no node specified
        connected = True

    if not connected:
      print(Fore.RED + 'Could not connect to target {} with user {}'.format(self.args.node, self.args.user)) 
      print("Is the target '{}' up and running and reachable?\nIs the user '{}' allowed to connect to the target via SSH?".format(self.args.node, self.args.user))
      if usecredentials:
        print("Is the password correct?")
      if usekeyfile:
        print("Is the key file '{}' correct and allowed on the target?".format(self.args.sshkey_file))
        print("You can add your key to the target with 'ssh-copy-id -i {} {}@{}'".format(self.args.sshkey_file, self.args.user, self.args.node))
        print("Alternatively you can specify a password using the '-p' or '--password' parameter if your target accepts login with password.")
      exit(1)

      #raise Exception('Could not connect to target') 
    for hw in self.targethardwarelist:
      try:  
        if self.cmd(hw['hardware']['runcheck']).strip() == hw['hardware']['checkresult']:
          self.target = hw
          print("{} detected".format(self.target['fullname']))
          break
      except: 
        pass
    if not self.target:
      if self.remote_client != None:
        self.remote_client.close()
      print('Could not detect target hardware. If you want to use a remote node use the --node parameter to specify the IP address.')
      #raise Exception('Could not detect target hardware. If you want to use a remote node use the --node parameter to specify the IP address.')
      exit(1)
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    if self.remote_client != None:
      self.remote_client.close()

  def connect_with_key(self, user, key_filename):
    if 'node' in self.args and self.args.node != None:
      print('trying to connect to {} with user {} using keyfile {}...'.format(self.args.node, user, key_filename), end='')
      client = paramiko.SSHClient()
      client.load_system_host_keys()
      #client.set_missing_host_key_policy(paramiko.WarningPolicy())
      client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
      client.connect(self.args.node, username=user, key_filename=key_filename)
      self.remote_client = client
      print(Fore.GREEN + 'succeeded')
    return self

  def connect(self, user, password):
    if 'node' in self.args and self.args.node != None:
      print('trying to connect to {} with user {}...'.format(self.args.node, user), end='')
      client = paramiko.SSHClient()
      client.load_system_host_keys()
      #client.set_missing_host_key_policy(paramiko.WarningPolicy())
      client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
      client.connect(self.args.node, username=user, password=password)
      self.remote_client = client
      print(Fore.GREEN + 'succeeded')
    return self
    
  def need_sudo_pw(self):
    try:
      if self.cmd_retcode('sudo -n true') == 0:
        return False
    except:
      pass
    return True

  def cmd(self, command, throw_on_nonzero_retcode=False):
    if not self.remote_client:
      p = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                          )
      output, err = p.communicate()
      output = output.decode('utf-8', 'ignore')
      if throw_on_nonzero_retcode and p.returncode != 0:
        raise Exception('Cound not execute {} \nError: {}'.format(command, err))

      return output
    else:
      ret = -1
      chan = self.remote_client.get_transport().open_session()
      chan.exec_command(command)
      try:
        stdout = chan.makefile('rb').read().decode('ascii', 'replace')
        stderr = chan.makefile_stderr('rb').read().decode('ascii', 'replace')
      except Exception as e: 
        print(str(e))
      ret = chan.recv_exit_status()
      chan.close()
      if throw_on_nonzero_retcode and ret != 0:
        raise Exception('Cound not execute {} \nError: {}'.format(command, stderr))
      return stdout

  def cmd_retcode(self, command):
    if not self.remote_client:
      return subprocess.run(command, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE).returncode
    else:
      ret = -1
      chan = self.remote_client.get_transport().open_session()
      chan.exec_command(command)
      ret = chan.recv_exit_status()
      chan.close()
      return ret

  def reboot(self):
    return self.cmd_retcode('sudo reboot -f > /dev/null 2>&1 &')

  def getFileContent(self, file):
    return self.cmd('sudo cat {}'.format(file), True)

  def setFileContent(self, file, content):
    cmdstr = 'sudo mkdir -p "{}" && printf "{}" | sudo tee {} >/dev/null'.format(os.path.dirname(file), content.replace('"', '\\"'), file)
    return self.cmd(cmdstr, True)

  def stop_service(self, name, verbose = True):
    trystop = False

    servicestarted = self.cmd_retcode('sudo systemctl is-active --quiet ' + name)
    if servicestarted == 0:
      if verbose:
        print(name + ' running, stopping...', end='')
        servicestopped = self.cmd_retcode('sudo service ' + name + ' stop')
        trystop = True # also set to true if stopping was unsucessful
        if verbose:
          if servicestopped == 0:
            print(Fore.GREEN + 'done')
          else:
            print(Fore.RED + 'failed')
    return trystop


  def start_service(self, name, verbose = True):
    if verbose:
      print('starting ' + name + '...', end='')
    servicestarted = self.cmd_retcode('sudo service ' + name + ' start')
    if verbose:
      if servicestarted == 0:
        print(Fore.GREEN + 'done')
      else:
        print(Fore.RED + 'failed')


  def check_hardware(self):
    if not 'hardware' in self.args:
      try:
        for hw in targethardwarelist:
          self.sbc.cmd(hw['hardware']['runcheck'])
      except:
        return True #for now do nothing on error, assume that it is a raspberry


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pilot_api.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pilot_api.proto',
  package='pilot_api',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0fpilot_api.proto\x12\tpilot_api\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1egoogle/protobuf/wrappers.proto\"2\n\x0c\x44\x65\x62ugModeMsg\x12\"\n\x04mode\x18\x01 \x01(\x0e\x32\x14.pilot_api.DebugMode\"\xc1\x01\n\nMetricInfo\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x15\n\rfriendly_name\x18\x03 \x01(\t\x12\x10\n\x08location\x18\x04 \x01(\t\x12*\n\x0bmetric_type\x18\x05 \x01(\x0e\x32\x15.pilot_api.MetricType\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x11\n\tprecision\x18\x07 \x01(\r\x12\x0c\n\x04unit\x18\x08 \x01(\t\x12\r\n\x05group\x18\t \x01(\t\"\"\n\x10\x46wConfigResponse\x12\x0e\n\x06\x63onfig\x18\x01 \x01(\t\"(\n\nIdResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\t\"\x1b\n\x0bUserRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x9c\x01\n\tMetricMsg\x12\x0e\n\x04\x62ool\x18\x01 \x01(\x08H\x00\x12\x10\n\x06\x64ouble\x18\x02 \x01(\x01H\x00\x12\x0f\n\x05\x66loat\x18\x03 \x01(\x02H\x00\x12\x0f\n\x05int32\x18\x04 \x01(\x11H\x00\x12\x0f\n\x05int64\x18\x05 \x01(\x12H\x00\x12\x10\n\x06uint32\x18\x06 \x01(\rH\x00\x12\x10\n\x06uint64\x18\x07 \x01(\x04H\x00\x12\r\n\x03str\x18\x08 \x01(\tH\x00\x42\x07\n\x05value\"q\n\x0f\x46irmwareRequest\x12)\n\x06target\x18\x01 \x01(\x0e\x32\x19.pilot_api.FirmwareTarget\x12\x0e\n\x06\x62inary\x18\x02 \x01(\x0c\x12\x11\n\tvariables\x18\x03 \x01(\t\x12\x10\n\x08\x66wconfig\x18\x04 \x01(\t\"\"\n\x0e\x46irmwareStatus\x12\x10\n\x08progress\x18\x01 \x01(\r*9\n\tDebugMode\x12\x0c\n\x08\x44\x65\x62ugOff\x10\x00\x12\x10\n\x0c\x44\x65\x62ugModules\x10\x01\x12\x0c\n\x08\x44\x65\x62ugAll\x10\x02* \n\nMetricType\x12\x06\n\x02IO\x10\x00\x12\n\n\x06Sensor\x10\x01*4\n\x0e\x46irmwareTarget\x12\x12\n\x0ePilotMainboard\x10\x00\x12\x0e\n\nSmartPower\x10\x01\x32\xed\x04\n\x08PilotApi\x12\x42\n\x0bGetFwConfig\x12\x16.google.protobuf.Empty\x1a\x1b.pilot_api.FwConfigResponse\x12\x36\n\x05GetID\x12\x16.google.protobuf.Empty\x1a\x15.pilot_api.IdResponse\x12\x43\n\x07SetName\x12\x1c.google.protobuf.StringValue\x1a\x1a.google.protobuf.BoolValue\x12\x43\n\x0cSetDebugMode\x12\x17.pilot_api.DebugModeMsg\x1a\x1a.google.protobuf.BoolValue\x12?\n\x0cGetDebugMode\x12\x16.google.protobuf.Empty\x1a\x17.pilot_api.DebugModeMsg\x12@\n\rGetMetricInfo\x12\x16.google.protobuf.Empty\x1a\x15.pilot_api.MetricInfo0\x01\x12M\n\x0eGetFloatMetric\x12\x1c.google.protobuf.StringValue\x1a\x1b.google.protobuf.FloatValue0\x01\x12\x41\n\tGetMetric\x12\x1c.google.protobuf.StringValue\x1a\x14.pilot_api.MetricMsg0\x01\x12\x46\n\x0bSetFirmware\x12\x1a.pilot_api.FirmwareRequest\x1a\x19.pilot_api.FirmwareStatus0\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,])

_DEBUGMODE = _descriptor.EnumDescriptor(
  name='DebugMode',
  full_name='pilot_api.DebugMode',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DebugOff', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DebugModules', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DebugAll', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=756,
  serialized_end=813,
)
_sym_db.RegisterEnumDescriptor(_DEBUGMODE)

DebugMode = enum_type_wrapper.EnumTypeWrapper(_DEBUGMODE)
_METRICTYPE = _descriptor.EnumDescriptor(
  name='MetricType',
  full_name='pilot_api.MetricType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='IO', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='Sensor', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=815,
  serialized_end=847,
)
_sym_db.RegisterEnumDescriptor(_METRICTYPE)

MetricType = enum_type_wrapper.EnumTypeWrapper(_METRICTYPE)
_FIRMWARETARGET = _descriptor.EnumDescriptor(
  name='FirmwareTarget',
  full_name='pilot_api.FirmwareTarget',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PilotMainboard', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SmartPower', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=849,
  serialized_end=901,
)
_sym_db.RegisterEnumDescriptor(_FIRMWARETARGET)

FirmwareTarget = enum_type_wrapper.EnumTypeWrapper(_FIRMWARETARGET)
DebugOff = 0
DebugModules = 1
DebugAll = 2
IO = 0
Sensor = 1
PilotMainboard = 0
SmartPower = 1



_DEBUGMODEMSG = _descriptor.Descriptor(
  name='DebugModeMsg',
  full_name='pilot_api.DebugModeMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='mode', full_name='pilot_api.DebugModeMsg.mode', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=141,
)


_METRICINFO = _descriptor.Descriptor(
  name='MetricInfo',
  full_name='pilot_api.MetricInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pilot_api.MetricInfo.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='pilot_api.MetricInfo.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='friendly_name', full_name='pilot_api.MetricInfo.friendly_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='location', full_name='pilot_api.MetricInfo.location', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metric_type', full_name='pilot_api.MetricInfo.metric_type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='pilot_api.MetricInfo.description', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='pilot_api.MetricInfo.precision', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unit', full_name='pilot_api.MetricInfo.unit', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group', full_name='pilot_api.MetricInfo.group', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=144,
  serialized_end=337,
)


_FWCONFIGRESPONSE = _descriptor.Descriptor(
  name='FwConfigResponse',
  full_name='pilot_api.FwConfigResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='pilot_api.FwConfigResponse.config', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=339,
  serialized_end=373,
)


_IDRESPONSE = _descriptor.Descriptor(
  name='IdResponse',
  full_name='pilot_api.IdResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pilot_api.IdResponse.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uuid', full_name='pilot_api.IdResponse.uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=375,
  serialized_end=415,
)


_USERREQUEST = _descriptor.Descriptor(
  name='UserRequest',
  full_name='pilot_api.UserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pilot_api.UserRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=417,
  serialized_end=444,
)


_METRICMSG = _descriptor.Descriptor(
  name='MetricMsg',
  full_name='pilot_api.MetricMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bool', full_name='pilot_api.MetricMsg.bool', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double', full_name='pilot_api.MetricMsg.double', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='float', full_name='pilot_api.MetricMsg.float', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int32', full_name='pilot_api.MetricMsg.int32', index=3,
      number=4, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int64', full_name='pilot_api.MetricMsg.int64', index=4,
      number=5, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uint32', full_name='pilot_api.MetricMsg.uint32', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uint64', full_name='pilot_api.MetricMsg.uint64', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='str', full_name='pilot_api.MetricMsg.str', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='pilot_api.MetricMsg.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=447,
  serialized_end=603,
)


_FIRMWAREREQUEST = _descriptor.Descriptor(
  name='FirmwareRequest',
  full_name='pilot_api.FirmwareRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='target', full_name='pilot_api.FirmwareRequest.target', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='binary', full_name='pilot_api.FirmwareRequest.binary', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='variables', full_name='pilot_api.FirmwareRequest.variables', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fwconfig', full_name='pilot_api.FirmwareRequest.fwconfig', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=605,
  serialized_end=718,
)


_FIRMWARESTATUS = _descriptor.Descriptor(
  name='FirmwareStatus',
  full_name='pilot_api.FirmwareStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='progress', full_name='pilot_api.FirmwareStatus.progress', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=720,
  serialized_end=754,
)

_DEBUGMODEMSG.fields_by_name['mode'].enum_type = _DEBUGMODE
_METRICINFO.fields_by_name['metric_type'].enum_type = _METRICTYPE
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['bool'])
_METRICMSG.fields_by_name['bool'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['double'])
_METRICMSG.fields_by_name['double'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['float'])
_METRICMSG.fields_by_name['float'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['int32'])
_METRICMSG.fields_by_name['int32'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['int64'])
_METRICMSG.fields_by_name['int64'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['uint32'])
_METRICMSG.fields_by_name['uint32'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['uint64'])
_METRICMSG.fields_by_name['uint64'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_METRICMSG.oneofs_by_name['value'].fields.append(
  _METRICMSG.fields_by_name['str'])
_METRICMSG.fields_by_name['str'].containing_oneof = _METRICMSG.oneofs_by_name['value']
_FIRMWAREREQUEST.fields_by_name['target'].enum_type = _FIRMWARETARGET
DESCRIPTOR.message_types_by_name['DebugModeMsg'] = _DEBUGMODEMSG
DESCRIPTOR.message_types_by_name['MetricInfo'] = _METRICINFO
DESCRIPTOR.message_types_by_name['FwConfigResponse'] = _FWCONFIGRESPONSE
DESCRIPTOR.message_types_by_name['IdResponse'] = _IDRESPONSE
DESCRIPTOR.message_types_by_name['UserRequest'] = _USERREQUEST
DESCRIPTOR.message_types_by_name['MetricMsg'] = _METRICMSG
DESCRIPTOR.message_types_by_name['FirmwareRequest'] = _FIRMWAREREQUEST
DESCRIPTOR.message_types_by_name['FirmwareStatus'] = _FIRMWARESTATUS
DESCRIPTOR.enum_types_by_name['DebugMode'] = _DEBUGMODE
DESCRIPTOR.enum_types_by_name['MetricType'] = _METRICTYPE
DESCRIPTOR.enum_types_by_name['FirmwareTarget'] = _FIRMWARETARGET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DebugModeMsg = _reflection.GeneratedProtocolMessageType('DebugModeMsg', (_message.Message,), {
  'DESCRIPTOR' : _DEBUGMODEMSG,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.DebugModeMsg)
  })
_sym_db.RegisterMessage(DebugModeMsg)

MetricInfo = _reflection.GeneratedProtocolMessageType('MetricInfo', (_message.Message,), {
  'DESCRIPTOR' : _METRICINFO,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.MetricInfo)
  })
_sym_db.RegisterMessage(MetricInfo)

FwConfigResponse = _reflection.GeneratedProtocolMessageType('FwConfigResponse', (_message.Message,), {
  'DESCRIPTOR' : _FWCONFIGRESPONSE,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.FwConfigResponse)
  })
_sym_db.RegisterMessage(FwConfigResponse)

IdResponse = _reflection.GeneratedProtocolMessageType('IdResponse', (_message.Message,), {
  'DESCRIPTOR' : _IDRESPONSE,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.IdResponse)
  })
_sym_db.RegisterMessage(IdResponse)

UserRequest = _reflection.GeneratedProtocolMessageType('UserRequest', (_message.Message,), {
  'DESCRIPTOR' : _USERREQUEST,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.UserRequest)
  })
_sym_db.RegisterMessage(UserRequest)

MetricMsg = _reflection.GeneratedProtocolMessageType('MetricMsg', (_message.Message,), {
  'DESCRIPTOR' : _METRICMSG,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.MetricMsg)
  })
_sym_db.RegisterMessage(MetricMsg)

FirmwareRequest = _reflection.GeneratedProtocolMessageType('FirmwareRequest', (_message.Message,), {
  'DESCRIPTOR' : _FIRMWAREREQUEST,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.FirmwareRequest)
  })
_sym_db.RegisterMessage(FirmwareRequest)

FirmwareStatus = _reflection.GeneratedProtocolMessageType('FirmwareStatus', (_message.Message,), {
  'DESCRIPTOR' : _FIRMWARESTATUS,
  '__module__' : 'pilot_api_pb2'
  # @@protoc_insertion_point(class_scope:pilot_api.FirmwareStatus)
  })
_sym_db.RegisterMessage(FirmwareStatus)



_PILOTAPI = _descriptor.ServiceDescriptor(
  name='PilotApi',
  full_name='pilot_api.PilotApi',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=904,
  serialized_end=1525,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetFwConfig',
    full_name='pilot_api.PilotApi.GetFwConfig',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_FWCONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetID',
    full_name='pilot_api.PilotApi.GetID',
    index=1,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_IDRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetName',
    full_name='pilot_api.PilotApi.SetName',
    index=2,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetDebugMode',
    full_name='pilot_api.PilotApi.SetDebugMode',
    index=3,
    containing_service=None,
    input_type=_DEBUGMODEMSG,
    output_type=google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDebugMode',
    full_name='pilot_api.PilotApi.GetDebugMode',
    index=4,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_DEBUGMODEMSG,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetMetricInfo',
    full_name='pilot_api.PilotApi.GetMetricInfo',
    index=5,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_METRICINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetFloatMetric',
    full_name='pilot_api.PilotApi.GetFloatMetric',
    index=6,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=google_dot_protobuf_dot_wrappers__pb2._FLOATVALUE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetMetric',
    full_name='pilot_api.PilotApi.GetMetric',
    index=7,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=_METRICMSG,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetFirmware',
    full_name='pilot_api.PilotApi.SetFirmware',
    index=8,
    containing_service=None,
    input_type=_FIRMWAREREQUEST,
    output_type=_FIRMWARESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_PILOTAPI)

DESCRIPTOR.services_by_name['PilotApi'] = _PILOTAPI

# @@protoc_insertion_point(module_scope)
